# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import copy
from collections.abc import MutableMapping, Sequence

import awkward as ak
from awkward._backends.backend import Backend
from awkward._layout import maybe_posaxis
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import ArrayLike, IndexType, NumpyMetadata
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._parameters import (
    parameters_intersect,
    parameters_union,
    type_parameters_equal,
)
from awkward._regularize import is_integer, is_integer_like
from awkward._slicing import NO_HEAD
from awkward._typing import TYPE_CHECKING, Callable, Final, Self, SupportsIndex, final
from awkward._util import UNSET
from awkward.contents.content import Content
from awkward.forms.form import Form
from awkward.forms.regularform import RegularForm
from awkward.index import Index

if TYPE_CHECKING:
    from awkward._slicing import SliceItem
    from awkward.contents.listoffsetarray import ListOffsetArray

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@final
class RegularArray(Content):
    """
    RegularArray describes lists that all have the same length, the single
    integer `size`. Its underlying `content` is a flattened view of the data;
    that is, each list is not stored separately in memory, but is inferred as a
    subinterval of the underlying data.

    If the `content` length is not an integer multiple of `size`, then the length
    of the RegularArray is truncated to the largest integer multiple.

    An extra field `zeros_length` is ignored unless the `size` is zero. This sets the
    length of the RegularArray in only those cases, so that it is possible for an
    array to contain a non-zero number of zero-length lists with regular type.

    A multidimensional #ak.contents.NumpyArray is equivalent to a one-dimensional
    #ak.layout.NumpyArray nested within several RegularArrays, one for each
    dimension. However, RegularArrays can be used to make lists of any other type.

    Like #ak.contents.ListArray and #ak.contents.ListOffsetArray, a RegularArray can
    represent strings if its `__array__` parameter is `"string"` (UTF-8 assumed) or
    `"bytestring"` (no encoding assumed) and it contains an #ak.contents.NumpyArray
    of `dtype=np.uint8` whose `__array__` parameter is `"char"` (UTF-8 assumed) or
    `"byte"` (no encoding assumed).

    RegularArray corresponds to an Apache Arrow
    [FixedSizeList](https://arrow.apache.org/docs/format/Columnar.html#fixed-size-list-layout).

    To illustrate how the constructor arguments are interpreted, the following is a
    simplified implementation of `__init__`, `__len__`, and `__getitem__`:

        class RegularArray(Content):
            def __init__(self, content, size, zeros_length=0):
                assert isinstance(content, Content)
                assert isinstance(size, int)
                assert isinstance(zeros_length, int)
                assert size >= 0
                if size != 0:
                    length = len(content) // size  # floor division
                else:
                    assert zeros_length >= 0
                    length = zeros_length
                self.content = content
                self.size = size
                self.length = length

            def __len__(self):
                return self.length

            def __getitem__(self, where):
                if isinstance(where, int):
                    if where < 0:
                        where += len(self)
                    assert 0 <= where < len(self)
                    return self.content[(where) * self.size : (where + 1) * self.size]

                elif isinstance(where, slice) and where.step is None:
                    start = where.start * self.size
                    stop = where.stop * self.size
                    zeros_length = where.stop - where.start
                    return RegularArray(
                        self.content[start:stop], self.size, zeros_length
                    )

                elif isinstance(where, str):
                    return RegularArray(self.content[where], self.size, self.length)

                else:
                    raise AssertionError(where)
    """

    is_list = True
    is_regular = True

    def __init__(self, content, size, zeros_length=0, *, parameters=None):
        if not isinstance(content, Content):
            raise TypeError(
                "{} 'content' must be a Content subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if size is unknown_length:
            if content.backend.index_nplike.known_data:
                raise TypeError(
                    "{} 'size' must be a non-negative integer for backends with known shapes, not None".format(
                        type(self).__name__
                    )
                )
        else:
            if not (is_integer(size) and size >= 0):
                raise TypeError(
                    "{} 'size' must be a non-negative integer, not {}".format(
                        type(self).__name__, size
                    )
                )

        if zeros_length is unknown_length:
            if content.backend.index_nplike.known_data:
                raise TypeError(
                    "{} 'zeros_length' must be a non-negative integer for backends with known shapes, not None".format(
                        type(self).__name__
                    )
                )
        else:
            if not (is_integer(zeros_length) and zeros_length >= 0):
                raise TypeError(
                    "{} 'zeros_length' must be a non-negative integer, not {}".format(
                        type(self).__name__, zeros_length
                    )
                )

        if parameters is not None and parameters.get("__array__") == "string":
            if not content.is_numpy or not content.parameter("__array__") == "char":
                raise ValueError(
                    "{} is a string, so its 'content' must be uint8 NumpyArray of char, not {}".format(
                        type(self).__name__, repr(content)
                    )
                )
        if parameters is not None and parameters.get("__array__") == "bytestring":
            if not content.is_numpy or not content.parameter("__array__") == "byte":
                raise ValueError(
                    "{} is a bytestring, so its 'content' must be uint8 NumpyArray of byte, not {}".format(
                        type(self).__name__, repr(content)
                    )
                )

        self._content = content
        self._size = size
        if content.length is unknown_length or size is unknown_length:
            self._length = unknown_length
        elif size != 0:
            self._length = content.length // size  # floor division
        else:
            self._length = zeros_length
        self._init(parameters, content.backend)

    @property
    def content(self) -> Content:
        return self._content

    @property
    def size(self):
        return self._size

    form_cls: Final = RegularForm

    def copy(self, content=UNSET, size=UNSET, zeros_length=UNSET, *, parameters=UNSET):
        return RegularArray(
            self._content if content is UNSET else content,
            self._size if size is UNSET else size,
            self._length if zeros_length is UNSET else zeros_length,
            parameters=self._parameters if parameters is UNSET else parameters,
        )

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy(
            content=copy.deepcopy(self._content, memo),
            parameters=copy.deepcopy(self._parameters, memo),
        )

    @classmethod
    def simplified(cls, content, size, zeros_length=0, *, parameters=None):
        return cls(content, size, zeros_length, parameters=parameters)

    @property
    def offsets(self) -> Index:
        return self._compact_offsets64(True)

    @property
    def starts(self) -> Index:
        return self._compact_offsets64(True)[:-1]

    @property
    def stops(self):
        return self._compact_offsets64(True)[1:]

    def _form_with_key(self, getkey: Callable[[Content], str | None]) -> RegularForm:
        form_key = getkey(self)
        return self.form_cls(
            self._content._form_with_key(getkey),
            self._size,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(
        self,
        form: Form,
        getkey: Callable[[Content, Form, str], str],
        container: MutableMapping[str, ArrayLike],
        backend: Backend,
        byteorder: str,
    ):
        assert isinstance(form, self.form_cls)
        self._content._to_buffers(form.content, getkey, container, backend, byteorder)

    def _to_typetracer(self, forget_length: bool) -> Self:
        return RegularArray(
            self._content._to_typetracer(forget_length),
            self._size,
            unknown_length if forget_length else self._length,
            parameters=self._parameters,
        )

    def _touch_data(self, recursive: bool):
        if recursive:
            self._content._touch_data(recursive)

    def _touch_shape(self, recursive: bool):
        if recursive:
            self._content._touch_shape(recursive)

    @property
    def length(self) -> ShapeItem:
        return self._length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<RegularArray size="]
        out.append(repr(str(self._size)))
        out.append(" len=")
        out.append(repr(str(self._length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</RegularArray>")
        out.append(post)
        return "".join(out)

    def to_ListOffsetArray64(self, start_at_zero: bool = False) -> ListOffsetArray:
        offsets = self._compact_offsets64(start_at_zero)
        return self._broadcast_tooffsets64(offsets)

    def to_RegularArray(self):
        return self

    def maybe_to_NumpyArray(self) -> ak.contents.NumpyArray | None:
        content = self._content[: self._length * self._size].maybe_to_NumpyArray()

        if content is not None:
            shape = (self._length, self._size) + content.data.shape[1:]
            return ak.contents.NumpyArray(
                self._backend.nplike.reshape(content.data, shape),
                parameters=parameters_union(self._parameters, content.parameters),
                backend=content.backend,
            )
        else:
            return None

    def _getitem_nothing(self):
        return self._content._getitem_range(0, 0)

    def _getitem_at(self, where: IndexType):
        index_nplike = self._backend.index_nplike
        where = index_nplike.regularize_index_for_length(where, self._length)
        size_scalar = index_nplike.shape_item_as_index(self._size)
        start, stop = where * size_scalar, (where + 1) * size_scalar
        return self._content._getitem_range(start, stop)

    def _getitem_range(self, start: SupportsIndex, stop: IndexType) -> Content:
        index_nplike = self._backend.index_nplike
        if not index_nplike.known_data:
            self._touch_shape(recursive=False)
            return self

        zeros_length = index_nplike.index_as_shape_item(stop - start)
        substart, substop = start * index_nplike.shape_item_as_index(
            self._size
        ), stop * index_nplike.shape_item_as_index(self._size)
        return RegularArray(
            self._content._getitem_range(substart, substop),
            self._size,
            zeros_length,
            parameters=self._parameters,
        )

    def _getitem_field(
        self, where: str | SupportsIndex, only_fields: tuple[str, ...] = ()
    ) -> Content:
        return RegularArray(
            self._content._getitem_field(where, only_fields),
            self._size,
            self._length,
            parameters=None,
        )

    def _getitem_fields(
        self, where: list[str | SupportsIndex], only_fields: tuple[str, ...] = ()
    ) -> Content:
        return RegularArray(
            self._content._getitem_fields(where, only_fields),
            self._size,
            self._length,
            parameters=None,
        )

    def _carry(self, carry: Index, allow_lazy: bool) -> Content:
        assert isinstance(carry, ak.index.Index)

        where = carry.data

        copied = allow_lazy == "copied"
        if not issubclass(where.dtype.type, np.int64):
            where = self._backend.index_nplike.astype(where, dtype=np.int64)
            copied = True

        negative = where < 0
        if self._backend.index_nplike.known_data:
            if self._backend.index_nplike.any(negative):
                if not copied:
                    where = where.copy()
                    copied = True
                where[negative] += self._length

            if self._backend.index_nplike.any(where >= self._length):
                raise ak._errors.index_error(self, where)

        if where.shape[0] is unknown_length or self._size is unknown_length:
            nextcarry = ak.index.Index64.empty(
                unknown_length, self._backend.index_nplike
            )
        else:
            nextcarry = ak.index.Index64.empty(
                where.shape[0] * self._size, self._backend.index_nplike
            )

        assert nextcarry.nplike is self._backend.index_nplike
        self._maybe_index_error(
            self._backend[
                "awkward_RegularArray_getitem_carry",
                nextcarry.dtype.type,
                where.dtype.type,
            ](
                nextcarry.data,
                where,
                where.shape[0],
                self._size,
            ),
            slicer=carry,
        )

        return RegularArray(
            self._content._carry(nextcarry, allow_lazy),
            self._size,
            where.shape[0],
            parameters=self._parameters,
        )

    def _compact_offsets64(self, start_at_zero):
        index_nplike = self._backend.index_nplike
        if self._size is not unknown_length and self._size == 0:
            return ak.index.Index64.zeros(self._length + 1, nplike=index_nplike)
        else:
            return ak.index.Index64(
                index_nplike.arange(
                    0,
                    index_nplike.shape_item_as_index(self._length * self._size) + 1,
                    index_nplike.shape_item_as_index(self._size),
                    dtype=np.int64,
                ),
                nplike=index_nplike,
            )

    def _broadcast_tooffsets64(self, offsets: Index) -> ListOffsetArray:
        self._touch_data(recursive=False)
        offsets._touch_data()

        index_nplike = self._backend.index_nplike
        assert offsets.nplike is index_nplike
        if offsets.length is not unknown_length and offsets.length == 0:
            raise AssertionError(
                "broadcast_tooffsets64 can only be used with non-empty offsets"
            )
        elif index_nplike.known_data and offsets[0] != 0:
            raise AssertionError(
                f"broadcast_tooffsets64 can only be used with offsets that start at 0, not {offsets[0]}"
            )
        elif (
            offsets.length is not unknown_length
            and self._length is not unknown_length
            and offsets.length - 1 != self._length
        ):
            raise AssertionError(
                "cannot broadcast RegularArray of length {} to length {}".format(
                    self._length, offsets.length - 1
                )
            )

        if self._size is not unknown_length and self._size == 1:
            count = offsets.data[1:] - offsets.data[:-1]
            # Sanity check that our kernel isn't losing values here
            assert (
                not self._backend.index_nplike.known_data
                or count.size is unknown_length
                or count.size == 0
                or count.dtype == np.intp
                or self._backend.index_nplike.max(count) <= np.iinfo(np.intp).max
            )
            carry = ak.index.Index64(
                index_nplike.repeat(
                    index_nplike.arange(
                        index_nplike.shape_item_as_index(self._length), dtype=np.int64
                    ),
                    index_nplike.astype(count, np.intp),
                ),
                nplike=index_nplike,
            )
            next_content = self._content._carry(carry, True)
        else:
            this_offsets = self._compact_offsets64(True)
            if index_nplike.known_data and not index_nplike.array_equal(
                offsets.data, this_offsets.data
            ):
                raise ValueError("cannot broadcast nested list")

            next_content = self._content[: offsets[-1]]
        return ak.contents.ListOffsetArray(
            offsets, next_content, parameters=self._parameters
        )

    def _getitem_next_jagged(
        self, slicestarts: Index, slicestops: Index, slicecontent: Content, tail
    ) -> Content:
        out = self.to_ListOffsetArray64(True)
        return out._getitem_next_jagged(slicestarts, slicestops, slicecontent, tail)

    def _getitem_next(
        self,
        head: SliceItem | tuple,
        tail: tuple[SliceItem, ...],
        advanced: Index | None,
    ) -> Content:
        index_nplike = self._backend.index_nplike

        if head is NO_HEAD:
            return self

        elif is_integer_like(head):
            nexthead, nexttail = ak._slicing.head_tail(tail)
            nextcarry = ak.index.Index64.empty(self._length, index_nplike)
            assert nextcarry.nplike is index_nplike
            self._maybe_index_error(
                self._backend[
                    "awkward_RegularArray_getitem_next_at", nextcarry.dtype.type
                ](
                    nextcarry.data,
                    head,
                    self._length,
                    self._size,
                ),
                slicer=head,
            )
            nextcontent = self._content._carry(nextcarry, True)
            return nextcontent._getitem_next(nexthead, nexttail, advanced)

        elif isinstance(head, slice):
            nexthead, nexttail = ak._slicing.head_tail(tail)
            start, stop, step, nextsize = index_nplike.derive_slice_for_length(
                head, length=self._size
            )

            nextcarry = ak.index.Index64.empty(self._length * nextsize, index_nplike)
            assert nextcarry.nplike is index_nplike
            self._maybe_index_error(
                self._backend[
                    "awkward_RegularArray_getitem_next_range",
                    nextcarry.dtype.type,
                ](
                    nextcarry.data,
                    start,
                    step,
                    self._length,
                    self._size,
                    nextsize,
                ),
                slicer=head,
            )

            nextcontent = self._content._carry(nextcarry, True)

            if advanced is None or (
                advanced.length is not unknown_length and advanced.length == 0
            ):
                return RegularArray(
                    nextcontent._getitem_next(nexthead, nexttail, advanced),
                    nextsize,
                    self._length,
                    parameters=self._parameters,
                )
            else:
                nextadvanced = ak.index.Index64.empty(nextcarry.length, index_nplike)
                advanced = advanced.to_nplike(index_nplike)
                assert (
                    nextadvanced.nplike is index_nplike
                    and advanced.nplike is index_nplike
                )
                self._maybe_index_error(
                    self._backend[
                        "awkward_RegularArray_getitem_next_range_spreadadvanced",
                        nextadvanced.dtype.type,
                        advanced.dtype.type,
                    ](
                        nextadvanced.data,
                        advanced.data,
                        self._length,
                        nextsize,
                    ),
                    slicer=head,
                )
                return RegularArray(
                    nextcontent._getitem_next(nexthead, nexttail, nextadvanced),
                    nextsize,
                    self._length,
                    parameters=self._parameters,
                )

        elif isinstance(head, str):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak.index.Index64):
            head = head.to_nplike(index_nplike)
            flathead = index_nplike.reshape(index_nplike.asarray(head.data), (-1,))
            regular_flathead = ak.index.Index64.empty(flathead.shape[0], index_nplike)
            assert regular_flathead.nplike is index_nplike
            self._maybe_index_error(
                self._backend[
                    "awkward_RegularArray_getitem_next_array_regularize",
                    regular_flathead.dtype.type,
                    flathead.dtype.type,
                ](
                    regular_flathead.data,
                    flathead,
                    flathead.shape[0],
                    self._size,
                ),
                slicer=head,
            )

            nexthead, nexttail = ak._slicing.head_tail(tail)
            if advanced is None or (
                advanced.length is not unknown_length and advanced.length == 0
            ):
                nextcarry = ak.index.Index64.empty(
                    self._length * flathead.shape[0],
                    index_nplike,
                )
                nextadvanced = ak.index.Index64.empty(nextcarry.length, index_nplike)
                assert (
                    nextcarry.nplike is index_nplike
                    and nextadvanced.nplike is index_nplike
                    and regular_flathead.nplike is index_nplike
                )
                self._maybe_index_error(
                    self._backend[
                        "awkward_RegularArray_getitem_next_array",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        regular_flathead.dtype.type,
                    ](
                        nextcarry.data,
                        nextadvanced.data,
                        regular_flathead.data,
                        self._length,
                        regular_flathead.length,
                        self._size,
                    ),
                    slicer=head,
                )
                nextcontent = self._content._carry(nextcarry, True)

                out = nextcontent._getitem_next(nexthead, nexttail, nextadvanced)
                if advanced is None:
                    return ak._slicing.getitem_next_array_wrap(
                        out, head.metadata.get("shape", (head.length,)), self._length
                    )
                else:
                    return out

            elif self._size == 0:
                nextcarry = ak.index.Index64.empty(0, nplike=index_nplike)
                nextadvanced = ak.index.Index64.empty(0, nplike=index_nplike)
                nextcontent = self._content._carry(nextcarry, True)
                return nextcontent._getitem_next(nexthead, nexttail, nextadvanced)

            else:
                nextcarry = ak.index.Index64.empty(self._length, index_nplike)
                nextadvanced = ak.index.Index64.empty(self._length, index_nplike)
                advanced = advanced.to_nplike(index_nplike)
                assert (
                    nextcarry.nplike is index_nplike
                    and nextadvanced.nplike is index_nplike
                    and advanced.nplike is index_nplike
                    and regular_flathead.nplike is index_nplike
                )
                self._maybe_index_error(
                    self._backend[
                        "awkward_RegularArray_getitem_next_array_advanced",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        advanced.dtype.type,
                        regular_flathead.dtype.type,
                    ](
                        nextcarry.data,
                        nextadvanced.data,
                        advanced.data,
                        regular_flathead.data,
                        self._length,
                        regular_flathead.length,
                        self._size,
                    ),
                    slicer=head,
                )
                nextcontent = self._content._carry(nextcarry, True)
                return nextcontent._getitem_next(nexthead, nexttail, nextadvanced)

        elif isinstance(head, ak.contents.ListOffsetArray):
            headlength = head.length
            head = head.to_backend(self._backend)

            if advanced is not None:
                raise ak._errors.index_error(
                    self,
                    head,
                    "cannot mix jagged slice with NumPy-style advanced indexing",
                )

            if self._backend.nplike.known_data and head.length != self._size:
                raise ak._errors.index_error(
                    self,
                    head,
                    "cannot fit jagged slice with length {} into {} of size {}".format(
                        head.length, type(self).__name__, self._size
                    ),
                )

            multistarts = ak.index.Index64.empty(
                head.length * self._length, index_nplike
            )
            multistops = ak.index.Index64.empty(
                head.length * self._length, index_nplike
            )

            assert head.offsets.nplike is index_nplike
            self._maybe_index_error(
                self._backend[
                    "awkward_RegularArray_getitem_jagged_expand",
                    multistarts.dtype.type,
                    multistops.dtype.type,
                    head.offsets.dtype.type,
                ](
                    multistarts.data,
                    multistops.data,
                    head.offsets.data,
                    head.length,
                    self._length,
                ),
                slicer=head,
            )
            down = self._content._getitem_next_jagged(
                multistarts, multistops, head._content, tail
            )

            return RegularArray(
                down, headlength, self._length, parameters=self._parameters
            )

        elif isinstance(head, ak.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _offsets_and_flattened(self, axis: int, depth: int) -> tuple[Index, Content]:
        return self.to_ListOffsetArray64(True)._offsets_and_flattened(axis, depth)

    def _mergeable_next(self, other: Content, mergebool: bool) -> bool:
        # Is the other content is an identity, or a union?
        if other.is_identity_like or other.is_union:
            return True
        # Check against option contents
        elif other.is_option or other.is_indexed:
            return self._mergeable_next(other.content, mergebool)
        # Otherwise, do the parameters match? If not, we can't merge.
        elif not type_parameters_equal(self._parameters, other._parameters):
            return False
        elif isinstance(
            other,
            (
                ak.contents.RegularArray,
                ak.contents.ListArray,
                ak.contents.ListOffsetArray,
            ),
        ):
            return self._content._mergeable_next(other.content, mergebool)
        elif isinstance(other, ak.contents.NumpyArray) and len(other.shape) > 1:
            return self._mergeable_next(other._to_regular_primitive(), mergebool)
        else:
            return False

    def _mergemany(self, others: Sequence[Content]) -> Content:
        if len(others) == 0:
            return self

        if any(x.is_option for x in others):
            return ak.contents.UnmaskedArray(self)._mergemany(others)

        # Regularize NumpyArray into RegularArray (or NumpyArray if 1D)
        others = [
            o.to_RegularArray() if isinstance(o, ak.contents.NumpyArray) else o
            for o in others
        ]

        if all(x.is_regular and x.size == self.size for x in others):
            parameters = self._parameters
            tail_contents = []
            zeros_length = self._length
            for x in others:
                parameters = parameters_intersect(parameters, x._parameters)
                tail_contents.append(x._content[: x._length * x._size])
                zeros_length += x._length

            return RegularArray(
                self._content[: self._length * self._size]._mergemany(tail_contents),
                self._size,
                zeros_length,
                parameters=parameters,
            )

        else:
            return self.to_ListOffsetArray64(True)._mergemany(others)

    def _fill_none(self, value: Content) -> Content:
        return RegularArray(
            self._content._fill_none(value),
            self._size,
            self._length,
            parameters=self._parameters,
        )

    def _local_index(self, axis, depth):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._local_index_axis0()
        elif posaxis is not None and posaxis + 1 == depth + 1:
            localindex = ak.index.Index64.empty(
                self._length * self._size, nplike=self._backend.index_nplike
            )
            self._backend.maybe_kernel_error(
                self._backend["awkward_RegularArray_localindex", np.int64](
                    localindex.data,
                    self._size,
                    self._length,
                )
            )
            return ak.contents.RegularArray(
                ak.contents.NumpyArray(localindex.data), self._size, self._length
            )
        else:
            return ak.contents.RegularArray(
                self._content._local_index(axis, depth + 1), self._size, self._length
            )

    def _numbers_to_type(self, name, including_unknown):
        return ak.contents.RegularArray(
            self._content._numbers_to_type(name, including_unknown),
            self._size,
            self._length,
            parameters=self._parameters,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._length == 0:
            return True

        return self.to_ListOffsetArray64(True)._is_unique(
            negaxis,
            starts,
            parents,
            outlength,
        )

    def _unique(self, negaxis, starts, parents, outlength):
        if self._length == 0:
            return self
        out = self.to_ListOffsetArray64(True)._unique(
            negaxis,
            starts,
            parents,
            outlength,
        )

        if isinstance(out, ak.contents.RegularArray):
            if isinstance(out._content, ak.contents.ListOffsetArray):
                return ak.contents.RegularArray(
                    out._content.to_RegularArray(),
                    out._size,
                    out._length,
                    parameters=out._parameters,
                )

        return out

    def _argsort_next(
        self, negaxis, starts, shifts, parents, outlength, ascending, stable
    ):
        next = self.to_ListOffsetArray64(True)
        out = next._argsort_next(
            negaxis, starts, shifts, parents, outlength, ascending, stable
        )

        if isinstance(out, ak.contents.RegularArray):
            if isinstance(out._content, ak.contents.ListOffsetArray):
                return ak.contents.RegularArray(
                    out._content.to_RegularArray(),
                    out._size,
                    out._length,
                    parameters=out._parameters,
                )

        return out

    def _sort_next(self, negaxis, starts, parents, outlength, ascending, stable):
        out = self.to_ListOffsetArray64(True)._sort_next(
            negaxis, starts, parents, outlength, ascending, stable
        )

        # FIXME
        # if isinstance(out, ak.contents.RegularArray):
        #     if isinstance(out._content, ak.contents.ListOffsetArray):
        #         return ak.contents.RegularArray(
        #             out._content.to_RegularArray(),
        #             out._size,
        #             out._length,
        #             None,
        #             out._parameters,
        #             self._backend.nplike,
        #         )

        return out

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        index_nplike = self._backend.index_nplike

        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        elif posaxis is not None and posaxis + 1 == depth + 1:
            if (
                self.parameter("__array__") == "string"
                or self.parameter("__array__") == "bytestring"
            ):
                raise ValueError(
                    "ak.combinations does not compute combinations of the characters of a string; please split it into lists"
                )

            size = self._size
            if replacement:
                size = size + (n - 1)
            thisn = n
            if thisn > size:
                combinationslen = 0
            elif thisn == size:
                combinationslen = 1
            else:
                if thisn * 2 > size:
                    thisn = size - thisn
                combinationslen = size
                for j in range(2, thisn + 1):
                    combinationslen = combinationslen * (size - j + 1)
                    combinationslen = combinationslen // j

            totallen = combinationslen * self._length
            tocarryraw = index_nplike.empty(n, dtype=np.intp)
            tocarry = []
            for i in range(n):
                ptr = ak.index.Index64.empty(
                    totallen,
                    nplike=index_nplike,
                    dtype=np.int64,
                )
                tocarry.append(ptr)
                if self._backend.nplike.known_data:
                    tocarryraw[i] = ptr.ptr

            toindex = ak.index.Index64.empty(n, index_nplike, dtype=np.int64)
            fromindex = ak.index.Index64.empty(n, index_nplike, dtype=np.int64)

            if self._size != 0:
                assert (
                    toindex.nplike is index_nplike and fromindex.nplike is index_nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_RegularArray_combinations_64",
                        np.int64,
                        toindex.data.dtype.type,
                        fromindex.data.dtype.type,
                    ](
                        tocarryraw,
                        toindex.data,
                        fromindex.data,
                        n,
                        replacement,
                        self._size,
                        self._length,
                    )
                )

            contents = []
            length = UNSET
            for ptr in tocarry:
                contents.append(self._content._carry(ptr, True))
                length = contents[-1].length
            assert length is not UNSET
            recordarray = ak.contents.RecordArray(
                contents,
                recordlookup,
                length,
                parameters=parameters,
                backend=self._backend,
            )
            return ak.contents.RegularArray(
                recordarray, combinationslen, self._length, parameters=self._parameters
            )
        else:
            length = self._length * self._size
            next = self._content._getitem_range(
                0, index_nplike.shape_item_as_index(length)
            )._combinations(n, replacement, recordlookup, parameters, axis, depth + 1)
            return ak.contents.RegularArray(
                next, self._size, self._length, parameters=self._parameters
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
        behavior,
    ):
        index_nplike = self._backend.index_nplike
        branch, depth = self.branch_depth
        nextlen = self._length * self._size
        if not branch and negaxis == depth:
            nextcarry = ak.index.Index64.empty(nextlen, nplike=index_nplike)
            nextparents = ak.index.Index64.empty(nextlen, nplike=index_nplike)
            assert (
                parents.nplike is index_nplike
                and nextcarry.nplike is index_nplike
                and nextparents.nplike is index_nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_RegularArray_reduce_nonlocal_preparenext",
                    nextcarry.dtype.type,
                    nextparents.dtype.type,
                    parents.dtype.type,
                ](
                    nextcarry.data,
                    nextparents.data,
                    parents.data,
                    self._size,
                    self._length,
                )
            )
            nextstarts = ak.index.Index64.empty(
                # `starts` must have at least enough elements for the largest `nextparent` to index into
                # The upper bound for this value is given by `nextlen` (each item in this list belonging
                # to a distinct reduction), but the length of `starts` should equate to `maxnextparents - 1`.
                starts.length * self._size,
                nplike=index_nplike,
            )
            assert (
                nextstarts.nplike is index_nplike and nextparents.nplike is index_nplike
            )
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_ListOffsetArray_reduce_nonlocal_nextstarts_64",
                    nextstarts.dtype.type,
                    nextparents.dtype.type,
                ](
                    nextstarts.data,
                    nextparents.data,
                    nextlen,
                )
            )

            if reducer.needs_position:
                # Regular arrays have the same length rows, so there can be no "missing" values
                # unlike ragged list types
                nextshifts = ak.index.Index64.zeros(
                    nextcarry.length, nplike=index_nplike
                )
            else:
                nextshifts = None

            nextcontent = self._content._carry(nextcarry, False)
            nextoutlength = outlength * self._size
            outcontent = nextcontent._reduce_next(
                reducer,
                negaxis - 1,
                nextstarts,
                nextshifts,
                nextparents,
                # We want a result of length
                nextoutlength,
                mask,
                False,
                behavior,
            )

            out = ak.contents.RegularArray(
                outcontent, self._size, outlength, parameters=None
            )

            if keepdims:
                out = ak.contents.RegularArray(out, 1, self._length, parameters=None)
            return out
        else:
            nextparents = ak.index.Index64.empty(nextlen, index_nplike)

            assert nextparents.nplike is index_nplike
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_RegularArray_reduce_local_nextparents",
                    nextparents.dtype.type,
                ](
                    nextparents.data,
                    self._size,
                    self._length,
                )
            )

            if self._size > 0:
                nextstarts = ak.index.Index64(
                    index_nplike.arange(0, nextlen, self._size),
                    nplike=index_nplike,
                )
            else:
                assert nextlen == 0
                nextstarts = ak.index.Index64(
                    index_nplike.zeros(nextlen, dtype=np.int64),
                    nplike=index_nplike,
                )

            outcontent = self._content._reduce_next(
                reducer,
                negaxis,
                nextstarts,
                shifts,
                nextparents,
                self._length,
                mask,
                keepdims,
                behavior,
            )

            # Understanding this logic is nontrivial without examples. So, let's start with a
            # somewhat simple example of `5 * 4 * 3 * int64`. This is composed of the following layouts
            #                                     ^........^ NumpyArray(5*4*3)
            #                                 ^............^ RegularArray(..., size=3)
            #                             ^................^ RegularArray(..., size=4)
            # If the reduction axis is 2 (or -1), then the resulting shape is `5 * 4 * int64`.
            # This corresponds to `RegularArray(NumpyArray(5*4), size=4)`, i.e. the `size=3` layout
            # *disappears*, and is replaced by the bare reduction `NumpyArray`. Therefore, if
            # this layout is _directly_ above the reduction, it won't survive the reduction.
            #
            # The `_reduce_next` mechanism returns its first result (by recursion) from the
            # leaf `NumpyArray`. The parent `RegularArray` takes the `negaxis != depth` branch,
            # and asks the `NumpyArray` to perform the reduction. The `_reduce_next` is such that the
            # callee must wrap the reduction result into a list whose length is given by the *caller*,
            # i.e. the parent. Therefore, the `RegularArray(..., size=3)` layout reduces its content with
            # `outlength=len(self)=5*4`. The parent of this `RegularArray` (with `size=4`)
            # will reduce its contents with `outlength=5`. It is *this* reduction that must be returned as a
            # `RegularArray(..., size=4)`. Clearly, the `RegularArray(..., size=3)` layout has disappeared
            # and been replaced by a `NumpyArray` of the appropriate size for the `RegularArray(..., size=4)`
            # to wrap. Hence, we only want to interpret the content as a `RegularArray(..., size=self._size)`
            # iff. we are not the direct parent of the reduced layout.

            # At `depth == negaxis+1`, we are above the dimension being reduced. All implemented
            # _dimensional_ types should return `RegularArray` for `keepdims=True`, so we don't need
            # to convert `outcontent` to a `RegularArray`
            if keepdims and depth == negaxis + 1:
                assert outcontent.is_regular
            # At `depth >= negaxis + 2`, we are wrapping at _least_ one other list type. This list-type
            # may return a ListOffsetArray, which we need to convert to a `RegularArray`. We know that
            # the result can be reinterpreted as a `RegularArray`, because it's not the immediate parent
            # of the reduction, so it must exist in the type.
            elif depth >= negaxis + 2:
                if outcontent.is_list:
                    # Let's only deal with ListOffsetArray
                    outcontent = outcontent.to_ListOffsetArray64(False)
                    # Fast-path to convert data that we know should be regular (avoid a kernel call)
                    start, stop = (
                        outcontent.offsets[0],
                        outcontent.offsets[outcontent.offsets.length - 1],
                    )

                    trimmed = outcontent.content._getitem_range(start, stop)
                    assert len(trimmed) == self._size * outcontent.length

                    outcontent = ak.contents.RegularArray(
                        trimmed,
                        size=self._size,
                        zeros_length=self._length,
                    )
                else:
                    assert outcontent.is_regular

            outoffsets = ak.index.Index64.empty(outlength + 1, index_nplike)
            assert outoffsets.nplike is index_nplike and parents.nplike is index_nplike
            self._backend.maybe_kernel_error(
                self._backend[
                    "awkward_ListOffsetArray_reduce_local_outoffsets_64",
                    outoffsets.dtype.type,
                    parents.dtype.type,
                ](
                    outoffsets.data,
                    parents.data,
                    parents.length,
                    outlength,
                )
            )

            return ak.contents.ListOffsetArray(outoffsets, outcontent, parameters=None)

    def _validity_error(self, path):
        if self.size < 0:
            return f"at {path} ({type(self)!r}): size < 0"

        return self._content._validity_error(path + ".content")

    def _nbytes_part(self):
        return self.content._nbytes_part()

    def _pad_none(self, target, axis, depth, clip):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._pad_none_axis0(target, clip)

        elif posaxis is not None and posaxis + 1 == depth + 1:
            if not clip:
                if target < self._size:
                    return self
                else:
                    return self._pad_none(target, axis, depth, True)

            else:
                index = ak.index.Index64.empty(
                    self._length * target,
                    self._backend.index_nplike,
                )
                assert index.nplike is self._backend.index_nplike
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_RegularArray_rpad_and_clip_axis1", index.dtype.type
                    ](index.data, target, self._size, self._length)
                )
                next = ak.contents.IndexedOptionArray.simplified(
                    index, self._content, parameters=self._parameters
                )
                return ak.contents.RegularArray(
                    next,
                    target,
                    self._length,
                    parameters=self._parameters,
                )

        else:
            return ak.contents.RegularArray(
                self._content._pad_none(target, axis, depth + 1, clip),
                self._size,
                self._length,
                parameters=self._parameters,
            )

    def _to_backend_array(self, allow_missing, backend):
        array_param = self.parameter("__array__")
        if array_param in {"bytestring", "string"}:
            # As our array-of-strings _may_ be empty, we should pass the dtype
            dtype = np.str_ if array_param == "string" else np.bytes_
            return backend.nplike.asarray(self.to_list(), dtype=dtype)
        else:
            out = self._content._to_backend_array(allow_missing, backend)
            shape = (self._length, self._size) + out.shape[1:]

            # ShapeItem is a defined type, but some nplikes don't map onto the entire space; e.g.
            # NumPy never has `None` shape items. We require that if a shape-item is used between nplikes
            # they both be the same "known-shape-ness".
            assert (
                self._backend.index_nplike.known_data == self._backend.nplike.known_data
            )
            return self._backend.nplike.reshape(
                out[
                    : self._backend.nplike.shape_item_as_index(
                        self._length * self._size
                    )
                ],
                shape,
            )

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        assert self._backend.nplike.known_data

        if self.parameter("__array__") == "string":
            return self.to_ListOffsetArray64(False)._to_arrow(
                pyarrow, mask_node, validbytes, length, options
            )

        is_bytestring = self.parameter("__array__") == "bytestring"

        akcontent = self._content[: self._length * self._size]

        if is_bytestring:
            assert isinstance(akcontent, ak.contents.NumpyArray)

            return pyarrow.Array.from_buffers(
                ak._connect.pyarrow.to_awkwardarrow_type(
                    pyarrow.binary(self._size),
                    options["extensionarray"],
                    options["record_is_scalar"],
                    mask_node,
                    self,
                ),
                self._length,
                [
                    ak._connect.pyarrow.to_validbits(validbytes),
                    pyarrow.py_buffer(akcontent._raw(numpy)),
                ],
            )

        else:
            paarray = akcontent._to_arrow(
                pyarrow, None, None, self._length * self._size, options
            )

            content_type = pyarrow.list_(paarray.type).value_field.with_nullable(
                akcontent.is_option
            )

            return pyarrow.Array.from_buffers(
                ak._connect.pyarrow.to_awkwardarrow_type(
                    pyarrow.list_(content_type, self._size),
                    options["extensionarray"],
                    options["record_is_scalar"],
                    mask_node,
                    self,
                ),
                self._length,
                [
                    ak._connect.pyarrow.to_validbits(validbytes),
                ],
                children=[paarray],
                null_count=ak._connect.pyarrow.to_null_count(
                    validbytes, options["count_nulls"]
                ),
            )

    def _remove_structure(self, backend, options):
        if (
            self.parameter("__array__") == "string"
            or self.parameter("__array__") == "bytestring"
        ):
            return [self]
        else:
            index_nplike = self._backend.index_nplike
            content = self._content[
                : index_nplike.shape_item_as_index(self._length * self._size)
            ]
            contents = content._remove_structure(backend, options)
            if options["keepdims"]:
                return [
                    RegularArray(c, size=c.length, parameters=self._parameters)
                    for c in contents
                ]
            else:
                return contents

    def _drop_none(self) -> Content:
        return self.to_ListOffsetArray64()._drop_none()

    def _recursively_apply(
        self, action, behavior, depth, depth_context, lateral_context, options
    ):
        if options["regular_to_jagged"]:
            return self.to_ListOffsetArray64(False)._recursively_apply(
                action, behavior, depth, depth_context, lateral_context, options
            )

        if self._backend.nplike.known_data:
            content = self._content[: self._length * self._size]
        else:
            self._touch_data(recursive=False)
            content = self._content

        if options["return_array"]:

            def continuation():
                return RegularArray(
                    content._recursively_apply(
                        action,
                        behavior,
                        depth + 1,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    self._size,
                    self._length,
                    parameters=self._parameters if options["keep_parameters"] else None,
                )

        else:

            def continuation():
                content._recursively_apply(
                    action,
                    behavior,
                    depth + 1,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            behavior=behavior,
            backend=self._backend,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def to_packed(self) -> Self:
        index_nplike = self._backend.index_nplike
        length = self._length * self._size
        if self._content.length == length:
            content = self._content.to_packed()
        else:
            content = self._content[
                : index_nplike.shape_item_as_index(length)
            ].to_packed()

        return RegularArray(
            content, self._size, self._length, parameters=self._parameters
        )

    def _to_list(self, behavior, json_conversions):
        if not self._backend.nplike.known_data:
            raise TypeError("cannot convert typetracer arrays to Python lists")

        if self.parameter("__array__") == "bytestring":
            convert_bytes = (
                None if json_conversions is None else json_conversions["convert_bytes"]
            )
            content = ak._util.tobytes(self._content.data)
            length, size = self._length, self._size
            out = [None] * length
            if convert_bytes is None:
                for i in range(length):
                    out[i] = content[(i) * size : (i + 1) * size]
            else:
                for i in range(length):
                    out[i] = convert_bytes(content[(i) * size : (i + 1) * size])
            return out

        elif self.parameter("__array__") == "string":
            data = self._content.data
            if hasattr(data, "tobytes"):

                def tostring(x):
                    return x.tobytes().decode(errors="surrogateescape")

            else:

                def tostring(x):
                    return x.tostring().decode(errors="surrogateescape")

            length, size = self._length, self._size
            out = [None] * length
            for i in range(length):
                out[i] = tostring(data[(i) * size : (i + 1) * size])
            return out

        else:
            out = self._to_list_custom(behavior, json_conversions)
            if out is not None:
                return out

            content = self._content._to_list(behavior, json_conversions)
            length, size = self._length, self._size
            out = [None] * length
            for i in range(length):
                out[i] = content[(i) * size : (i + 1) * size]
            return out

    def _to_backend(self, backend: Backend) -> Self:
        content = self._content.to_backend(backend)
        return RegularArray(
            content, self._size, zeros_length=self._length, parameters=self._parameters
        )

    def _is_equal_to(self, other, index_dtype, numpyarray):
        return self._size == other.size and self._content.is_equal_to(
            other._content, index_dtype, numpyarray
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test():
    array = ak.contents.NumpyArray(np.zeros((3, 0), dtype=np.int32))
    buffs = ak.operations.to_buffers(array)
    new_array = ak.operations.from_buffers(*buffs)

    assert to_list(new_array) == [[], [], []]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_simplified_indexed():
    numbers = ak.contents.IndexedArray(
        ak.index.Index64([0, 2, 3]),
        ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
    )
    records = ak.contents.RecordArray(
        [ak.contents.NumpyArray(np.array([4.0, 3.0, 1.0], dtype=np.int64))], ["x"]
    )
    result = ak.concatenate((numbers, records), highlevel=False)
    assert result.is_equal_to(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 0, 1, 1, 1]),
            ak.index.Index64([0, 2, 3, 0, 1, 2]),
            [
                ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
                ak.contents.RecordArray(
                    [ak.contents.NumpyArray(np.array([4.0, 3.0, 1.0], dtype=np.int64))],
                    ["x"],
                ),
            ],
        )
    )


def test_simplified_indexed_categorical():
    numbers = ak.contents.IndexedArray(
        ak.index.Index64([0, 2, 3]),
        ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
        parameters={"__array__": "categorical"},
    )
    records = ak.contents.RecordArray(
        [ak.contents.NumpyArray(np.array([4.0, 3.0, 1.0], dtype=np.int64))], ["x"]
    )
    result = ak.concatenate((numbers, records), highlevel=False)
    assert result.is_equal_to(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 0, 1, 1, 1]),
            ak.index.Index64([0, 1, 2, 0, 1, 2]),
            [
                ak.contents.IndexedArray(
                    ak.index.Index64([0, 2, 3]),
                    ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
                    parameters={"__array__": "categorical"},
                ),
                ak.contents.RecordArray(
                    [ak.contents.NumpyArray(np.array([4.0, 3.0, 1.0], dtype=np.int64))],
                    ["x"],
                ),
            ],
        )
    )


def test_merge_indexed():
    records = ak.contents.IndexedArray(
        ak.index.Index64([0, 2, 3]),
        ak.contents.RecordArray(
            [
                ak.contents.NumpyArray(
                    np.array([4.0, 3.0, 1.0, 9.0, 8.0, 7.0], dtype=np.int64)
                )
            ],
            ["x"],
            parameters={"inner": "bar", "drop": "this"},
        ),
        parameters={"outer": "foo", "ignore": "me"},
    )
    union = ak.contents.UnionArray(
        ak.index.Index8([0, 0, 0, 1, 1, 1]),
        ak.index.Index64([0, 1, 2, 0, 1, 2]),
        [
            ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
            ak.contents.RecordArray(
                [ak.contents.NumpyArray(np.array([4.0, 3.0, 1.0], dtype=np.int64))],
                ["x"],
                parameters={"outer": "foo", "inner": "bar"},
            ),
        ],
    )
    result = ak.concatenate((union, records), highlevel=False)

    assert result.is_equal_to(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 0, 1, 1, 1, 1, 1, 1]),
            ak.index.Index64([0, 1, 2, 0, 1, 2, 3, 5, 6]),
            [
                ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
                ak.contents.RecordArray(
                    [
                        ak.contents.NumpyArray(
                            np.array(
                                [4.0, 3.0, 1.0, 4.0, 3.0, 1.0, 9.0, 8.0, 7.0],
                                dtype=np.int64,
                            )
                        )
                    ],
                    ["x"],
                ),
            ],
        )
    )
    # This test might be a bit strict; any code that views `layout.parameters` will change this result to `{}`
    assert result.contents[0]._parameters is None
    assert result.contents[1]._parameters == {"outer": "foo", "inner": "bar"}


def test_merge_indexed_categorical():
    records = ak.contents.IndexedArray(
        ak.index.Index64([0, 2, 3]),
        ak.contents.RecordArray(
            [
                ak.contents.NumpyArray(
                    np.array([4.0, 3.0, 1.0, 9.0, 8.0, 7.0], dtype=np.int64)
                )
            ],
            ["x"],
            parameters={"inner": "bar", "drop": "this"},
        ),
        parameters={"outer": "foo", "ignore": "me", "__array__": "categorical"},
    )
    union = ak.contents.UnionArray(
        ak.index.Index8([0, 0, 0, 1, 1, 1]),
        ak.index.Index64([0, 1, 2, 0, 1, 2]),
        [
            ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
            ak.contents.IndexedArray(
                ak.index.Index64([0, 1, 2]),
                ak.contents.RecordArray(
                    [ak.contents.NumpyArray(np.array([4.0, 3.0, 1.0], dtype=np.int64))],
                    ["x"],
                    parameters={"inner": "bar"},
                ),
                parameters={"outer": "foo", "__array__": "categorical"},
            ),
        ],
    )
    result = ak.concatenate((union, records), highlevel=False)
    assert result.is_equal_to(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 0, 1, 1, 1, 1, 1, 1]),
            ak.index.Index64([0, 1, 2, 0, 1, 2, 3, 4, 5]),
            [
                ak.contents.NumpyArray(np.arange(10, dtype=np.int64)),
                ak.contents.IndexedArray(
                    ak.index.Index64([0, 1, 2, 3, 5, 6]),
                    ak.contents.RecordArray(
                        [
                            ak.contents.NumpyArray(
                                np.array(
                                    [4.0, 3.0, 1.0, 4.0, 3.0, 1.0, 9.0, 8.0, 7.0],
                                    dtype=np.int64,
                                )
                            )
                        ],
                        ["x"],
                    ),
                    parameters={"__array__": "categorical"},
                ),
            ],
        )
    )
    # This test might be a bit strict; any code that views `layout.parameters` will change this result to `{}`
    assert result.contents[0]._parameters is None
    assert result.contents[1]._parameters == {
        "__array__": "categorical",
        "outer": "foo",
    }
    assert result.contents[1].content._parameters == {"inner": "bar"}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

try:
    import numba as nb
    import numba.cuda as nb_cuda
    from numba import config

    ak.numba.register_and_check()

except ImportError:
    nb = nb_cuda = None

numbatest = pytest.mark.skipif(
    nb is None or nb_cuda is None, reason="requires the numba and numba.cuda packages"
)

config.CUDA_LOW_OCCUPANCY_WARNINGS = False
config.CUDA_WARN_ON_IMPLICIT_COPY = False


@nb_cuda.jit(extensions=[ak.numba.cuda])
def multiply(array, n, out):
    tid = nb_cuda.grid(1)
    out[tid] = array[tid] * n


@nb_cuda.jit(extensions=[ak.numba.cuda])
def pass_through(array, out):
    x = nb_cuda.grid(1)
    out[x] = array[x] if array[x] is not None else np.nan


@nb_cuda.jit(extensions=[ak.numba.cuda])
def pass_through_2d(array, out):
    x, y = nb_cuda.grid(2)
    if x < len(array) and y < len(array[x]):
        out[x][y] = array[x][y]
    else:
        out[x][y] = np.nan


@nb_cuda.jit(extensions=[ak.numba.cuda])
def pass_regular_through_2d(array, out):
    x, y = nb_cuda.grid(2)
    out[x][y] = array[x][y]


@nb_cuda.jit(extensions=[ak.numba.cuda])
def pass_record_through(array, out):
    tid = nb_cuda.grid(1)
    out[tid] = array.x[tid]


@nb_cuda.jit(extensions=[ak.numba.cuda])
def count_records(array, out):
    tid = nb_cuda.grid(1)
    record = array[tid]
    out[tid] = np.nan if record is None else tid


@nb_cuda.jit(extensions=[ak.numba.cuda])
def pass_two_records_through(array, out):
    tid = nb_cuda.grid(1)
    out[tid][0] = array.x[tid]
    out[tid][1] = array.y[tid]


@nb_cuda.jit(extensions=[ak.numba.cuda])
def pass_two_tuple_through(array, out):
    tid = nb_cuda.grid(1)
    out[tid][0] = array["0"][tid]
    out[tid][1] = array["1"][tid]


@numbatest
def test_array_multiply():
    # create an ak.Array with a cuda backend:
    akarray = ak.Array([0, 1, 2, 3], backend="cuda")

    # allocate the result:
    results = nb_cuda.to_device(np.empty(4, dtype=np.int32))

    multiply[1, 4](akarray, 3, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == [0, 3, 6, 9]


@numbatest
def test_array_on_cpu_multiply():
    # create an ak.Array with a cpu backend:
    array = ak.Array([0, 1, 2, 3])

    # allocate the result:
    results = nb_cuda.to_device(np.empty(4, dtype=np.int32))

    with pytest.raises(TypeError):
        multiply[1, 4](array, 3, results)

    multiply[1, 4](ak.to_backend(array, backend="cuda"), 3, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == [0, 3, 6, 9]


@numbatest
def test_ListOffsetArray():
    array = ak.Array([[0, 1], [2], [3, 4, 5]], backend="cuda")

    results = nb_cuda.to_device(np.empty(9, dtype=np.float32).reshape((3, 3)))

    pass_through_2d[(3, 1), (1, 3)](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert (
        ak.drop_none(ak.nan_to_none(ak.Array(host_results))).tolist() == array.tolist()
    )


@numbatest
def test_RecordArray():
    array = ak.Array(
        [{"x": 0}, {"x": 1}, {"x": 2}, {"x": 3}, {"x": 4}, {"x": 5}], backend="cuda"
    )

    results = nb_cuda.to_device(np.empty(6, dtype=np.int32))

    pass_record_through[1, 6](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == array.x.to_list()


@numbatest
def test_EmptyArray():
    array = ak.Array(ak.contents.EmptyArray(), backend="cuda")

    results = nb_cuda.to_device(np.empty(len(array), dtype=np.int32))
    pass_through[1, 1](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == array.to_list()


@numbatest
def test_NumpyArray():
    array = ak.Array(ak.contents.NumpyArray([0, 1, 2, 3]), backend="cuda")

    results = nb_cuda.to_device(np.empty(len(array), dtype=np.int32))
    pass_through[1, 4](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == array.to_list()


@numbatest
def test_RegularArray_NumpyArray():
    array = ak.Array(
        ak.contents.RegularArray(
            ak.contents.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
            3,
        ),
        backend="cuda",
    )

    results = nb_cuda.to_device(np.empty(6, dtype=np.float64).reshape((2, 3)))
    pass_through_2d[(2, 1), (1, 3)](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == array.to_list()


@numbatest
def test_RegularArray_EmptyArray():
    array = ak.Array(
        ak.contents.RegularArray(ak.contents.EmptyArray(), 0, zeros_length=10),
        backend="cuda",
    )

    results = nb_cuda.to_device(np.empty(10, dtype=np.float64).reshape((10, 1)))
    pass_through_2d[(10, 1), (1, 1)](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert (
        ak.drop_none(ak.nan_to_none(ak.Array(host_results))).tolist() == array.to_list()
    )


@numbatest
def test_ListArray_NumpyArray():
    array = ak.Array(
        ak.contents.ListArray(
            ak.index.Index(np.array([4, 100, 1], np.int64)),
            ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
            ak.contents.NumpyArray(np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])),
        ),
        backend="cuda",
    )
    results = nb_cuda.to_device(np.empty(9, dtype=np.float64).reshape((3, 3)))

    pass_through_2d[(3, 1), (1, 3)](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.drop_none(ak.nan_to_none(ak.Array(host_results))).tolist() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]


@numbatest
def test_ListOffsetArray_NumpyArray():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index(np.array([1, 4, 4, 6, 7], np.int64)),
            ak.contents.NumpyArray(np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])),
        ),
        backend="cuda",
    )
    results = nb_cuda.to_device(np.empty(12, dtype=np.float64).reshape((4, 3)))

    pass_through_2d[(4, 1), (1, 3)](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert (
        ak.drop_none(ak.nan_to_none(ak.Array(host_results))).tolist() == array.to_list()
    )


@numbatest
def test_RecordArray_NumpyArray():
    array = ak.Array(
        ak.contents.RecordArray(
            [
                ak.contents.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
                ak.contents.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
            ],
            ["x", "y"],
        ),
        backend="cuda",
    )
    assert array.to_list() == [
        {"x": 0, "y": 0.0},
        {"x": 1, "y": 1.1},
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
        {"x": 4, "y": 4.4},
    ]
    results = nb_cuda.to_device(np.empty(10, dtype=np.float64).reshape((5, 2)))

    pass_two_records_through[5, 1](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.to_list(host_results) == [
        [0.0, 0.0],
        [1.0, 1.1],
        [2.0, 2.2],
        [3.0, 3.3],
        [4.0, 4.4],
    ]

    array = ak.Array(
        ak.contents.RecordArray(
            [
                ak.contents.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
                ak.contents.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
            ],
            None,
        ),
        backend="cuda",
    )
    assert array.to_list() == [(0, 0.0), (1, 1.1), (2, 2.2), (3, 3.3), (4, 4.4)]
    results = nb_cuda.to_device(np.empty(10, dtype=np.float64).reshape((5, 2)))

    pass_two_tuple_through[5, 1](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).to_list() == [
        [0.0, 0.0],
        [1.0, 1.1],
        [2.0, 2.2],
        [3.0, 3.3],
        [4.0, 4.4],
    ]

    array = ak.Array(
        ak.contents.RecordArray([], [], 10),
        backend="cuda",
    )
    assert array.to_list() == [{}, {}, {}, {}, {}, {}, {}, {}, {}, {}]
    results = nb_cuda.to_device(np.empty(10, dtype=np.int32))
    count_records[1, 10](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    array = ak.Array(
        ak.contents.RecordArray([], None, 10),
        backend="cuda",
    )
    assert array.to_list() == [(), (), (), (), (), (), (), (), (), ()]
    results = nb_cuda.to_device(np.empty(10, dtype=np.int32))
    count_records[1, 10](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


@numbatest
def test_IndexedArray_NumpyArray():
    array = ak.Array(
        ak.contents.IndexedArray(
            ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4], np.int64)),
            ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        ),
        backend="cuda",
    )
    results = nb_cuda.to_device(np.empty(7, dtype=np.float64))

    pass_through[1, 7](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.Array(host_results).tolist() == array.to_list()


@numbatest
def test_IndexedOptionArray_NumpyArray():
    array = ak.Array(
        ak.contents.IndexedOptionArray(
            ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4], np.int64)),
            ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        ),
        backend="cuda",
    )
    results = nb_cuda.to_device(np.empty(7, dtype=np.float64))

    pass_through[1, 7](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.nan_to_none(ak.Array(host_results)).tolist() == [
        3.3,
        3.3,
        None,
        2.2,
        None,
        6.6,
        5.5,
    ]


@numbatest
def test_ByteMaskedArray_NumpyArray():
    array = ak.Array(
        ak.contents.ByteMaskedArray(
            ak.index.Index(np.array([1, 0, 1, 0, 1], np.int8)),
            ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
            valid_when=True,
        ),
        backend="cuda",
    )
    results = nb_cuda.to_device(np.empty(5, dtype=np.float64))

    pass_through[1, 5](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.nan_to_none(ak.Array(host_results)).tolist() == array.to_list()

    array = ak.Array(
        ak.contents.ByteMaskedArray(
            ak.index.Index(np.array([0, 1, 0, 1, 0], np.int8)),
            ak.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
            valid_when=False,
        ),
        backend="cuda",
    )
    results = nb_cuda.to_device(np.empty(5, dtype=np.float64))

    pass_through[1, 5](array, results)

    nb_cuda.synchronize()
    host_results = results.copy_to_host()

    assert ak.nan_to_none(ak.Array(host_results)).tolist() == array.to_list()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak
from awkward._backends.numpy import NumpyBackend
from awkward._backends.typetracer import TypeTracerBackend


def test():
    layout = ak.to_layout({"x": 1, "y": 1})
    assert ak.backend(layout) == "cpu"
    assert layout.backend is NumpyBackend.instance()

    layout_tt = layout.to_backend("typetracer")
    assert ak.backend(layout_tt) == "typetracer"
    assert layout_tt.backend is TypeTracerBackend.instance()

    with pytest.raises(TypeError):
        layout_tt.to_backend("cpu")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_example():
    x = ak.operations.from_iter([[1, 2, 3, None], [], [4, 5]])
    y = ak.operations.from_iter([100, 200, 300])

    ttx = ak.highlevel.Array(x.layout.to_typetracer())
    tty = ak.highlevel.Array(y.layout.to_typetracer())

    assert (x + y).layout.form == (ttx + tty).layout.form
    assert (x + np.sin(y)).layout.form == (ttx + np.sin(tty)).layout.form

    x = ak.highlevel.Array(
        ak.contents.ListArray(x.layout.starts, x.layout.stops, x.layout.content)
    )
    ttx = ak.highlevel.Array(x.layout.to_typetracer())

    assert (x + x).layout.form == (ttx + ttx).layout.form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    layout = ak.contents.IndexedArray(
        ak.index.Index64(np.array([3, 1, 0, 2])),
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 3, 6, 9, 12])),
            ak.contents.NumpyArray(np.array([0, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 3])),
        ),
    )

    assert ak.operations.unflatten(
        layout,
        ak.operations.flatten(ak.operations.run_lengths(layout)),
        axis=1,
    ).to_list() == [[[3, 3, 3]], [[1, 1, 1]], [[0, 0, 0]], [[2, 2], [3]]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# Versioning
from awkward._version import __version__

# NumPy-like alternatives
import awkward._backends
import awkward._nplikes

# layout classes; functionality that used to be in C++ (in Awkward 1.x)
import awkward.index
import awkward.contents
import awkward.record
import awkward.types
import awkward.forms
import awkward._do
import awkward._slicing
import awkward._broadcasting
import awkward._reducers

# internal
import awkward._util
import awkward._errors
import awkward._lookup

# third-party connectors
import awkward._connect.numpy
import awkward._connect.numexpr
import awkward.numba
import awkward.cppyy
import awkward.jax
import awkward.typetracer
import awkward._typetracer  # todo: remove this after "deprecation" period

# high-level interface
from awkward.highlevel import *

# behaviors
import awkward.behaviors.categorical
import awkward.behaviors.string
from awkward.behaviors.mixins import *

# exports
import awkward.builder
import awkward.forth

behavior: dict = {}

# operations
from awkward.operations import *

# version
__all__ = [x for x in globals() if not x.startswith("_")]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    assert ak.almost_equal(
        ak.concatenate([ak.Array([1, 2, 3]), ak.Array([1, 2, None])]),
        ak.contents.ByteMaskedArray(
            ak.index.Index8(np.array([False, False, False, False, False, True])),
            ak.contents.NumpyArray(np.array([1, 2, 3, 1, 2, 3], dtype=np.int64)),
            valid_when=False,
        ),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_fix_longlong_type_passed_to_index_1530():
    a = np.asarray([0, 5, 10, 15, 20, 25], dtype=np.longlong)
    index = ak.index.Index64(a)
    content = ak.contents.NumpyArray(
        np.asarray(
            [101, 118, 116, 45, 48, 101, 118, 116, 45, 49, 101, 118], dtype=np.int64
        )
    )
    listoffsetarray = ak.contents.ListOffsetArray(index, content)

    assert ak.to_list(listoffsetarray) == [
        [101, 118, 116, 45, 48],
        [101, 118, 116, 45, 49],
        [101, 118],
        [],
        [],
    ]


def test_typetracer_view_method():
    a = np.asarray([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.longlong)
    b = ak.contents.NumpyArray(a)
    c = np.array([7, 3, 3, 5], dtype=np.longlong)
    e = ak.index.Index(b.to_typetracer()[c].data)
    f = ak.index.Index(a)

    assert e.form == f.form
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.to_typetracer()[c].form == b[c].form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


class Lookup:
    def __init__(self, layout, generator=None):
        positions = []
        tolookup(layout, positions)

        def arrayptr(x):
            if isinstance(x, int):
                return x
            elif isinstance(self.nplike, ak._nplikes.cupy.Cupy):
                return x.data.ptr
            else:
                return x.ctypes.data

        self.nplike = layout.backend.nplike
        self.generator = generator
        self.positions = positions
        self.arrayptrs = self.nplike.asarray(
            [arrayptr(x) for x in positions], dtype=np.intp
        )


def tolookup(layout, positions):
    if isinstance(layout, ak.contents.EmptyArray):
        return tolookup(layout.to_NumpyArray(np.dtype(np.float64)), positions)

    elif isinstance(layout, ak.contents.NumpyArray):
        if len(layout.shape) == 1:
            return NumpyLookup.tolookup(layout, positions)
        else:
            return tolookup(layout.to_RegularArray(), positions)

    elif isinstance(layout, ak.contents.RegularArray):
        return RegularLookup.tolookup(layout, positions)

    elif isinstance(layout, (ak.contents.ListArray, ak.contents.ListOffsetArray)):
        return ListLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.IndexedArray):
        return IndexedLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.IndexedOptionArray):
        return IndexedOptionLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.ByteMaskedArray):
        return ByteMaskedLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.BitMaskedArray):
        return BitMaskedLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.UnmaskedArray):
        return UnmaskedLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.RecordArray):
        return RecordLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.record.Record):
        return RecordLookup.tolookup(layout, positions)

    elif isinstance(layout, ak.contents.UnionArray):
        return UnionLookup.tolookup(layout, positions)

    else:
        raise AssertionError(f"unrecognized Content: {type(layout)}")


class ContentLookup:
    pass


class NumpyLookup(ContentLookup):
    ARRAY = 0

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(layout.to_contiguous().data)
        return pos

    def tolayout(self, lookup, pos, fields):
        assert fields == ()
        return ak.contents.NumpyArray(
            lookup.positions[pos + self.ARRAY], parameters=self.parameters
        )


class RegularLookup(ContentLookup):
    ZEROS_LENGTH = 0
    CONTENT = 1

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(len(layout))
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.RegularArray(
            content,
            self.size,
            lookup.positions[pos + self.ZEROS_LENGTH],
            parameters=self.parameters,
        )


class ListLookup(ContentLookup):
    STARTS = 0
    STOPS = 1
    CONTENT = 2

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(layout.starts.data)
        positions.append(layout.stops.data)
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        starts = self.IndexOf(self.indextype)(lookup.positions[pos + self.STARTS])
        stops = self.IndexOf(self.indextype)(lookup.positions[pos + self.STOPS])
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.ListArray(starts, stops, content, parameters=self.parameters)


class IndexedLookup(ContentLookup):
    INDEX = 0
    CONTENT = 1

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(layout.index.data)
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        index = self.IndexOf(self.indextype)(lookup.positions[pos + self.INDEX])
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.IndexedArray(index, content, parameters=self.parameters)


class IndexedOptionLookup(ContentLookup):
    INDEX = 0
    CONTENT = 1

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(layout.index.data)
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        index = self.IndexOf(self.indextype)(lookup.positions[pos + self.INDEX])
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.IndexedOptionArray(
            index, content, parameters=self.parameters
        )


class ByteMaskedLookup(ContentLookup):
    MASK = 0
    CONTENT = 1

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(layout.mask.data)
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        mask = self.IndexOf(self.masktype)(lookup.positions[pos + self.MASK])
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.ByteMaskedArray(
            mask, content, self.valid_when, parameters=self.parameters
        )


class BitMaskedLookup(ContentLookup):
    LENGTH = 0
    MASK = 1
    CONTENT = 2

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(len(layout))
        positions.append(layout.mask.data)
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        mask = self.IndexOf(self.masktype)(lookup.positions[pos + self.MASK])
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.BitMaskedArray(
            mask,
            content,
            self.valid_when,
            lookup.positions[pos + self.LENGTH],
            self.lsb_order,
            parameters=self.parameters,
        )


class UnmaskedLookup(ContentLookup):
    CONTENT = 0

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(None)
        positions[pos + cls.CONTENT] = tolookup(layout.content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        content = self.contenttype.tolayout(
            lookup, lookup.positions[pos + self.CONTENT], fields
        )
        return ak.contents.UnmaskedArray(content, parameters=self.parameters)


class RecordLookup(ContentLookup):
    LENGTH = 0
    CONTENTS = 1

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(len(layout))
        positions.extend([None] * len(layout.contents))
        for i, content in enumerate(layout.contents):
            positions[pos + cls.CONTENTS + i] = tolookup(content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        if len(fields) > 0:
            index = self.fieldindex(fields[0])
            assert index is not None
            return self.contenttypes[index].tolayout(
                lookup, lookup.positions[pos + self.CONTENTS + index], fields[1:]
            )
        else:
            contents = []
            for i, contenttype in enumerate(self.contenttypes):
                layout = contenttype.tolayout(
                    lookup, lookup.positions[pos + self.CONTENTS + i], fields
                )
                contents.append(layout)

            return ak.contents.RecordArray(
                contents,
                self.fields,
                lookup.positions[pos + self.LENGTH],
                parameters=self.parameters,
            )


class UnionLookup(ContentLookup):
    TAGS = 0
    INDEX = 1
    CONTENTS = 2

    @classmethod
    def tolookup(cls, layout, positions):
        pos = len(positions)
        positions.append(layout.tags.data)
        positions.append(layout.index.data)
        positions.extend([None] * len(layout.contents))
        for i, content in enumerate(layout.contents):
            positions[pos + cls.CONTENTS + i] = tolookup(content, positions)
        return pos

    def tolayout(self, lookup, pos, fields):
        tags = self.IndexOf(self.tagstype)(lookup.positions[pos + self.TAGS])
        index = self.IndexOf(self.indextype)(lookup.positions[pos + self.INDEX])
        contents = []
        for i, contenttype in enumerate(self.contenttypes):
            layout = contenttype.tolayout(
                lookup, lookup.positions[pos + self.CONTENTS + i], fields
            )
            contents.append(layout)
        return ak.contents.UnionArray(tags, index, contents, parameters=self.parameters)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.shape import unknown_length


def test():
    layout = ak.to_layout([[1, 2, 3, 4], [5, 6, 7, 8], [4, 5, 6, 9]]).to_typetracer(
        forget_length=True
    )
    result = layout.to_RegularArray()
    assert result.size is unknown_length
    assert result.length is unknown_length



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_broadcast_single_bool():
    base = ak.Array([[{"x": 0.1, "y": 0.2, "z": 0.3}, {"x": 0.4, "y": 0.5, "z": 0.6}]])
    base_new1 = ak.operations.with_field(base, True, "always_true")
    assert to_list(base_new1.always_true) == [[True, True]]
    base_new2 = ak.operations.with_field(base_new1, base.x > 0.3, "sometimes_true")
    assert to_list(base_new2.always_true) == [[True, True]]
    assert ak.operations.fields(base_new2) == [
        "x",
        "y",
        "z",
        "always_true",
        "sometimes_true",
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
import ctypes
import os
import textwrap

import cppyy
import ROOT

import awkward as ak
import awkward._connect.cling
import awkward._lookup
from awkward._backends.numpy import NumpyBackend
from awkward._layout import wrap_layout
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import NumpyMetadata
from awkward.types.numpytype import primitive_to_dtype

cpp_type_of = {
    "bool": "bool",
    "int8": "int8_t",
    "uint8": "uint8_t",
    "int16": "int16_t",
    "uint16": "uint16_t",
    "int32": "int32_t",
    "uint32": "uint32_t",
    "int64": "int64_t",
    "uint64": "uint64_t",
    "float32": "float",
    "float64": "double",
    "complex64": "std::complex<float>",
    "complex128": "std::complex<double>",
    "datetime64": "std::time_t",
    "timedelta64": "std::difftime",
}

np = NumpyMetadata.instance()
numpy = Numpy.instance()


cppyy.add_include_path(
    os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "include",
        )
    )
)
cppyy.add_include_path(
    os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            os.path.pardir,
            "header-only",
        )
    )
)


cppyy.include("rdataframe/jagged_builders.h")


def from_rdataframe(
    data_frame, columns, highlevel, behavior, with_name, offsets_type, keep_order
):
    def cpp_builder_type(depth, data_type):
        if depth == 1:
            return f"awkward::LayoutBuilder::Numpy<{data_type}>>"
        else:
            return f"awkward::LayoutBuilder::ListOffset<int64_t, {cpp_builder_type(depth - 1, data_type)}>"

    def cpp_fill_offsets_and_flatten(depth):
        if depth == 1:
            return textwrap.dedent(
                """
                for (auto const& it : vec1) {
                    builder1.append(it);
                }
                """
            )
        else:
            return textwrap.dedent(
                f"""
                for (auto const& vec{depth - 1} : vec{depth}) {{
                    auto& builder{depth - 1} = builder{depth}.begin_list();
                    {cpp_fill_offsets_and_flatten(depth - 1)}
                    builder{depth}.end_list();
                }}
                """
            )

    def cpp_fill_function(depth):
        if depth == 1:
            return textwrap.dedent(
                """
                template<class BUILDER, typename PRIMITIVE>
                void fill_from(BUILDER& builder, ROOT::RDF::RResultPtr<std::vector<PRIMITIVE>>& result) {
                    for (auto const& it : result) {
                        builder.append(it);
                    }
                }
                """
            )
        else:
            return textwrap.dedent(
                f"""
                template<class BUILDER, typename PRIMITIVE>
                void fill_offsets_and_flatten{depth}(BUILDER& builder{depth}, ROOT::RDF::RResultPtr<std::vector<PRIMITIVE>>& result) {{
                    for (auto const& vec{depth - 1} : result) {{
                      auto& builder{depth - 1} = builder{depth}.begin_list();
                      {cpp_fill_offsets_and_flatten(depth - 1)}
                      builder{depth}.end_list();
                    }}
                }}
                """
            )

    def form_dtype(form):
        if isinstance(form, ak.forms.NumpyForm) and form.inner_shape == ():
            return primitive_to_dtype(form.primitive)
        elif isinstance(form, ak.forms.ListOffsetForm):
            return form_dtype(form.content)

    # Register Take action for each column
    # 'Take' is a lazy action:
    column_types = {}
    result_ptrs = {}
    contents = {}
    index = {}
    awkward_type_cols = {}

    columns = (*columns, "rdfentry_")
    maybe_indexed = keep_order

    # Important note: This loop is separate from the next one
    # in order not to trigger the additional RDataFrame
    # Event loops
    for col in columns:
        column_types[col] = data_frame.GetColumnType(col)
        result_ptrs[col] = data_frame.Take[column_types[col]](col)

        if ROOT.awkward.is_awkward_type[column_types[col]]():
            maybe_indexed = True

    if not maybe_indexed:
        columns = columns[:-1]

    for col in columns:
        if ROOT.awkward.is_awkward_type[column_types[col]]():  # Retrieve Awkward arrays
            # ROOT::RDF::RResultPtr<T>::begin Returns an iterator to the beginning of
            # the contained object if this makes sense, throw a compilation error otherwise.
            #
            # Does not trigger event loop and execution of all actions booked in
            # the associated RLoopManager.
            lookup = result_ptrs[col].begin().lookup()
            generator = lookup[col].generator
            layout = generator.tolayout(lookup[col], 0, ())
            awkward_type_cols[col] = layout

        else:  # Convert the C++ vectors to Awkward arrays
            form_str = ROOT.awkward.type_to_form[column_types[col], offsets_type](0)

            if form_str == "unsupported type":
                raise TypeError(
                    f"{col!r} column's type {column_types[col]!r} is not supported."
                )

            form = ak.forms.from_json(form_str)

            list_depth = form.purelist_depth
            form_dtype_name = form_dtype(form).name
            data_type = cpp_type_of[form_dtype_name]

            # pull in the CppBuffers (after which we can import from it)
            CppBuffers = cppyy.gbl.awkward.CppBuffers
            cpp_buffers_self = CppBuffers()

            if isinstance(form, ak.forms.NumpyForm):
                NumpyBuilder = cppyy.gbl.awkward.LayoutBuilder.Numpy[data_type]
                builder = NumpyBuilder()
                builder_type = type(builder).__cpp_name__

                cpp_buffers_self.fill_from[builder_type, column_types[col]](
                    builder, result_ptrs[col]
                )

            elif isinstance(form, ak.forms.ListOffsetForm):
                if isinstance(form.content, ak.forms.NumpyForm):
                    # NOTE: list_depth == 2 or 1 if its the list of strings
                    list_depth = 2

                ListOffsetBuilder = cppyy.gbl.awkward.LayoutBuilder.ListOffset[
                    offsets_type,
                    cpp_builder_type(list_depth - 1, data_type),
                ]
                builder = ListOffsetBuilder()
                builder_type = type(builder).__cpp_name__

                if not hasattr(
                    cppyy.gbl.awkward, f"fill_offsets_and_flatten{list_depth}"
                ):
                    done = cppyy.cppdef(
                        "namespace awkward {" + cpp_fill_function(list_depth) + "}"
                    )
                    assert done is True

                fill_from_func = getattr(
                    cppyy.gbl.awkward, f"fill_offsets_and_flatten{list_depth}"
                )
                fill_from_func[builder_type, column_types[col]](
                    builder, result_ptrs[col]
                )
            else:
                raise AssertionError(f"unrecognized Form: {type(form)}")

            names_nbytes = cpp_buffers_self.names_nbytes[builder_type](builder)

            buffers = {}
            for item in names_nbytes:
                buffers[item.first] = numpy.empty(item.second, dtype=np.uint8)
                cpp_buffers_self.append(
                    item.first,
                    buffers[item.first].ctypes.data_as(ctypes.POINTER(ctypes.c_ubyte)),
                )

            length = cpp_buffers_self.to_char_buffers[builder_type](builder)

            if col == "rdfentry_":
                index[col] = ak.from_buffers(
                    form,
                    length,
                    buffers,
                    byteorder=ak._util.native_byteorder,
                    highlevel=highlevel,
                    behavior=behavior,
                )
                index[col] = ak.index.Index64(
                    index[col].layout.to_backend_array(
                        allow_missing=True, backend=NumpyBackend.instance()
                    )
                )
            else:
                contents[col] = ak.from_buffers(
                    form,
                    length,
                    buffers,
                    byteorder=ak._util.native_byteorder,
                    highlevel=highlevel,
                    behavior=behavior,
                )

    for key, value in awkward_type_cols.items():
        if len(index["rdfentry_"]) < len(value):
            contents[key] = wrap_layout(
                ak.contents.IndexedArray(index["rdfentry_"], value),
                highlevel=highlevel,
                behavior=behavior,
            )
        else:
            contents[key] = value

    out = ak.zip(
        contents,
        depth_limit=1,
        highlevel=highlevel,
        behavior=behavior,
        with_name=with_name,
    )

    if keep_order:
        sorted = ak.index.Index64(index["rdfentry_"].data.argsort())
        out = wrap_layout(
            ak.contents.IndexedArray(sorted, out.layout),
            highlevel=highlevel,
            behavior=behavior,
        )

    return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    layout = ak.contents.ByteMaskedArray(
        ak.index.Index8(np.array([0, 1, 0, 1, 0], dtype=np.int8)),
        ak.contents.NumpyArray(np.arange(5)),
        valid_when=True,
    )
    result = layout.to_ByteMaskedArray(False)
    assert layout.to_list() == [None, 1, None, 3, None]
    assert result.to_list() == [None, 1, None, 3, None]
    assert layout.backend.index_nplike.asarray(result.mask).tolist() == [1, 0, 1, 0, 1]

    # Check this works
    layout.to_typetracer().to_ByteMaskedArray(False)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
import pytest

import awkward as ak


def test_one_level_record():
    base = ak.zip(
        {"x": [1, 2, 3], "y": [8, 9, 10]},
    )

    assert ak.without_field(base, where=["x"]).to_list() == [
        {"y": 8},
        {"y": 9},
        {"y": 10},
    ]
    assert ak.fields(base) == ["x", "y"]

    del base["x"]
    assert base.to_list() == [
        {"y": 8},
        {"y": 9},
        {"y": 10},
    ]


def test_one_level_tuple():
    base = ak.zip(
        [[1, 2, 3], [8, 9, 10]],
    )

    assert ak.without_field(base, where=["0"]).to_list() == [
        (8,),
        (9,),
        (10,),
    ]
    assert ak.fields(base) == ["0", "1"]

    del base["0"]
    assert base.to_list() == [
        (8,),
        (9,),
        (10,),
    ]


def test_two_level_record():
    base = ak.zip(
        {"a": ak.zip({"x": [1, 2, 3], "y": [8, 9, 10]}), "b": [1, 2, 3]},
        depth_limit=1,
    )

    assert ak.without_field(base, where=["a", "x"]).to_list() == [
        {"b": 1, "a": {"y": 8}},
        {"b": 2, "a": {"y": 9}},
        {"b": 3, "a": {"y": 10}},
    ]
    assert ak.fields(base) == ["a", "b"]
    assert ak.fields(base["a"]) == ["x", "y"]

    del base["a", "x"]
    assert base.to_list() == [
        {"b": 1, "a": {"y": 8}},
        {"b": 2, "a": {"y": 9}},
        {"b": 3, "a": {"y": 10}},
    ]


def test_two_level_tuple():
    base = ak.zip(
        [ak.zip([[1, 2, 3], [8, 9, 10]]), [1, 2, 3]],
        depth_limit=1,
    )

    assert ak.without_field(base, where=["0", "0"]).to_list() == [
        ((8,), 1),
        ((9,), 2),
        ((10,), 3),
    ]
    assert ak.fields(base) == ["0", "1"]
    assert ak.fields(base["0"]) == ["0", "1"]

    del base["0", "0"]
    assert base.to_list() == [
        ((8,), 1),
        ((9,), 2),
        ((10,), 3),
    ]


def test_two_level_mixed():
    base = ak.zip(
        [ak.zip({"x": [1, 2, 3], "y": [8, 9, 10]}), [1, 2, 3]],
        depth_limit=1,
    )

    assert ak.without_field(base, where=["0", "x"]).to_list() == [
        ({"y": 8}, 1),
        ({"y": 9}, 2),
        ({"y": 10}, 3),
    ]
    assert ak.fields(base) == ["0", "1"]
    assert ak.fields(base["0"]) == ["x", "y"]

    del base["0", "x"]
    assert base.to_list() == [
        ({"y": 8}, 1),
        ({"y": 9}, 2),
        ({"y": 10}, 3),
    ]


def test_one_level_delete_the_only_field():
    base = ak.zip({"x": [1, 2, 3]})
    assert ak.without_field(base, where=["x"]).to_list() == [
        {},
        {},
        {},
    ]
    assert ak.fields(base) == ["x"]

    del base["x"]
    assert base.to_list() == [
        {},
        {},
        {},
    ]


def test_two_level_delete_the_only_field():
    base = ak.zip({"a": ak.zip({"x": [1, 2, 3]})}, depth_limit=1)
    assert ak.without_field(base, where=["a", "x"]).to_list() == [
        {"a": {}},
        {"a": {}},
        {"a": {}},
    ]
    assert ak.fields(base) == ["a"]

    del base["a", "x"]
    assert base.to_list() == [
        {"a": {}},
        {"a": {}},
        {"a": {}},
    ]


def test_check_no_fields():
    base = ak.zip({"a": ak.zip({"x": [1, 2, 3]})}, depth_limit=1)

    with pytest.raises(IndexError, match=r"no field"):
        ak.without_field(base, "x")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

array = ak.Array([[0, 2, 3.0], [4, 5, 6, 7, 8], [], [9, 8, None], [10, 1], []])


def test_sum():
    assert ak.sum(array, axis=None) == pytest.approx(63.0)
    assert ak.almost_equal(
        ak.sum(array, axis=None, keepdims=True), ak.to_regular([[63.0]])
    )
    assert ak.almost_equal(
        ak.sum(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[63.0]]).mask[[[True]]]),
    )
    assert ak.sum(array[2], axis=None, mask_identity=True) is None


def test_prod():
    assert ak.prod(array[1:], axis=None) == pytest.approx(4838400.0)
    assert ak.prod(array, axis=None) == 0
    assert ak.almost_equal(
        ak.prod(array, axis=None, keepdims=True), ak.to_regular([[0.0]])
    )
    assert ak.almost_equal(
        ak.prod(array[1:], axis=None, keepdims=True), ak.to_regular([[4838400.0]])
    )
    assert ak.almost_equal(
        ak.prod(array[1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[4838400.0]]).mask[[[True]]]),
    )
    assert ak.prod(array[2], axis=None, mask_identity=True) is None


def test_min():
    assert ak.min(array, axis=None) == pytest.approx(0.0)
    assert ak.almost_equal(
        ak.min(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[0.0]]),
    )
    assert ak.almost_equal(
        ak.min(array, axis=None, keepdims=True, initial=-100, mask_identity=False),
        ak.to_regular([[-100.0]]),
    )

    assert ak.almost_equal(
        ak.min(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0.0]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.min(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[np.inf]]).mask[[[False]]]),
    )
    assert ak.min(array[2], axis=None, mask_identity=True) is None


def test_max():
    assert ak.max(array, axis=None) == pytest.approx(10.0)
    assert ak.almost_equal(
        ak.max(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[10.0]]),
    )
    assert ak.almost_equal(
        ak.max(array, axis=None, keepdims=True, initial=100, mask_identity=False),
        ak.to_regular([[100.0]]),
    )
    assert ak.almost_equal(
        ak.max(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[10.0]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.max(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[np.inf]]).mask[[[False]]]),
    )
    assert ak.max(array[2], axis=None, mask_identity=True) is None


def test_count():
    assert ak.count(array, axis=None) == 12
    assert ak.almost_equal(
        ak.count(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[12]]),
    )
    assert ak.almost_equal(
        ak.count(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[12]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.count(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0]]).mask[[[False]]]),
    )
    assert ak.count(array[2], axis=None, mask_identity=True) is None
    assert ak.count(array[2], axis=None, mask_identity=False) == 0


def test_count_nonzero():
    assert ak.count_nonzero(array, axis=None) == 11
    assert ak.almost_equal(
        ak.count_nonzero(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[11]]),
    )
    assert ak.almost_equal(
        ak.count_nonzero(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[11]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.count_nonzero(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0]]).mask[[[False]]]),
    )
    assert ak.count_nonzero(array[2], axis=None, mask_identity=True) is None
    assert ak.count_nonzero(array[2], axis=None, mask_identity=False) == 0


def test_std():
    assert ak.std(array, axis=None) == pytest.approx(3.139134700306227)
    assert ak.almost_equal(
        ak.std(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[3.139134700306227]]),
    )
    assert ak.almost_equal(
        ak.std(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[3.139134700306227]]).mask[[[True]]]),
    )
    assert np.isnan(ak.std(array[2], axis=None, mask_identity=False))


def test_std_no_mask_axis_none():
    assert ak.almost_equal(
        ak.std(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0.0]]).mask[[[False]]]),
    )
    assert ak.std(array[2], axis=None, mask_identity=True) is None


def test_var():
    assert ak.var(array, axis=None) == pytest.approx(9.854166666666666)
    assert ak.almost_equal(
        ak.var(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[9.854166666666666]]),
    )
    assert ak.almost_equal(
        ak.var(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[9.854166666666666]]).mask[[[True]]]),
    )
    assert np.isnan(ak.var(array[2], axis=None, mask_identity=False))


def test_var_no_mask_axis_none():
    assert ak.almost_equal(
        ak.var(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0.0]]).mask[[[False]]]),
    )
    assert ak.var(array[2], axis=None, mask_identity=True) is None


def test_mean():
    assert ak.mean(array, axis=None) == pytest.approx(5.25)
    assert ak.almost_equal(
        ak.mean(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[5.25]]),
    )
    assert ak.almost_equal(
        ak.mean(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[5.25]]).mask[[[True]]]),
    )
    assert np.isnan(ak.mean(array[2], axis=None, mask_identity=False))


def test_mean_no_mask_axis_none():
    assert ak.almost_equal(
        ak.mean(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0.0]]).mask[[[False]]]),
    )
    assert ak.mean(array[2], axis=None, mask_identity=True) is None


def test_ptp():
    assert ak.ptp(array, axis=None) == pytest.approx(10.0)
    assert ak.almost_equal(
        ak.ptp(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[10.0]]),
    )
    assert ak.almost_equal(
        ak.ptp(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[10.0]]).mask[[[True]]]),
    )
    assert ak.ptp(array[2], axis=None, mask_identity=False) == pytest.approx(0.0)


def test_ptp_no_mask_axis_none():
    assert ak.almost_equal(
        ak.ptp(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0.0]]).mask[[[False]]]),
    )
    assert ak.ptp(array[2], axis=None, mask_identity=True) is None


def test_argmax():
    assert ak.argmax(array, axis=None) == 11
    assert ak.almost_equal(
        ak.argmax(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[11]]),
    )
    assert ak.almost_equal(
        ak.argmax(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[11]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.argmax(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0]]).mask[[[False]]]),
    )
    assert ak.argmax(array[2], axis=None, mask_identity=True) is None
    assert ak.argmax(array[2], axis=None, mask_identity=False) == -1


def test_argmin():
    assert ak.argmin(array, axis=None) == 0
    assert ak.almost_equal(
        ak.argmin(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[0]]),
    )
    assert ak.almost_equal(
        ak.argmin(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[0]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.argmin(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[999]]).mask[[[False]]]),
    )
    assert ak.argmin(array[2], axis=None, mask_identity=True) is None
    assert ak.argmin(array[2], axis=None, mask_identity=False) == -1


def test_any():
    assert ak.any(array, axis=None)
    assert ak.almost_equal(
        ak.any(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[True]]),
    )
    assert ak.almost_equal(
        ak.any(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[True]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.any(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[True]]).mask[[[False]]]),
    )
    assert ak.any(array[2], axis=None, mask_identity=True) is None
    assert not ak.any(array[2], axis=None, mask_identity=False)


def test_all():
    assert not ak.all(array, axis=None)
    assert ak.almost_equal(
        ak.all(array, axis=None, keepdims=True, mask_identity=False),
        ak.to_regular([[False]]),
    )
    assert ak.almost_equal(
        ak.all(array, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[False]]).mask[[[True]]]),
    )
    assert ak.almost_equal(
        ak.all(array[-1:], axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(ak.Array([[False]]).mask[[[False]]]),
    )
    assert ak.all(array[2], axis=None, mask_identity=True) is None
    assert ak.all(array[2], axis=None, mask_identity=False)


def test_prod_numpy():
    array_reg = ak.from_numpy(np.arange(7 * 5 * 3, dtype=np.int64).reshape((7, 5, 3)))
    result_reg = ak.from_numpy(np.array([[[5460]]], dtype=np.int64))

    assert ak.sum(array_reg, axis=None) == 5460
    assert ak.almost_equal(ak.sum(array_reg, axis=None, keepdims=True), result_reg)
    assert ak.almost_equal(
        ak.sum(array_reg, axis=None, keepdims=True, mask_identity=True),
        ak.to_regular(result_reg.mask[[[[True]]]], axis=None),
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.shape import unknown_length


def test():
    x = ak.to_regular(
        ak.Array([[[-1, 1.0], [-2, 2], [-3, 3]], [[-4, 4]], []], backend="typetracer"),
        axis=-1,
    )
    result = x[:, :-1]
    assert ak.backend(result) == "typetracer"
    layout = ak.to_layout(result)
    assert layout.length == 3
    assert layout.content.length is unknown_length
    assert layout.content.content.length is unknown_length
    assert layout.content.content.dtype == np.dtype(np.float64)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("with_name",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def with_name(array, name, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        name (str or None): Name to give to the records or tuples; this assigns
            the `"__record__"` parameter. If None, any existing name is unset.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an #ak.Array or #ak.Record (or low-level equivalent, if
    `highlevel=False`) with a new name. This function does not change the
    array in-place. If the new name is None, then an array without a name is
    returned.

    The records or tuples may be nested within multiple levels of nested lists.
    If records are nested within records, only the outermost are affected.

    Setting the `"__record__"` parameter makes it possible to add behaviors
    to the data; see #ak.Array and #ak.behavior for a more complete
    description.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, name, highlevel, behavior)


def _impl(array, name, highlevel, behavior):
    behavior = behavior_of(array, behavior=behavior)
    layout = ak.operations.to_layout(array)

    def action(layout, **ignore):
        if isinstance(layout, ak.contents.RecordArray):
            return ak.contents.RecordArray(
                layout._contents,
                layout._fields,
                layout._length,
                parameters={**layout.parameters, "__record__": name},
            )
        else:
            return None

    out = ak._do.recursively_apply(layout, action, behavior)

    def action2(layout, **ignore):
        if layout.is_union:
            return ak.contents.UnionArray.simplified(
                layout._tags,
                layout._index,
                layout._contents,
                parameters=layout._parameters,
            )
        else:
            return None

    out2 = ak._do.recursively_apply(out, action2, behavior)

    return wrap_layout(out2, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [
                {"x": 3.3, "y": [1, 2, None, 3]},
                False,
                False,
                True,
                {"x": 4.4, "y": [1, 2, None, 3, 4]},
            ],
        ]
    )

    def assert_array_type(new_array, intended_type):
        """helper function to check each part of the array took the intended type"""
        assert isinstance(new_array[0][0]["x"], intended_type)
        assert isinstance(new_array[0][1]["x"], intended_type)
        assert isinstance(new_array[0][1]["y"][0], intended_type)
        assert isinstance(new_array[0][2]["x"], intended_type)
        assert isinstance(new_array[0][2]["y"][0], intended_type)
        assert isinstance(new_array[0][2]["y"][1], intended_type)
        assert isinstance(new_array[2][0]["x"], intended_type)
        assert isinstance(new_array[2][0]["y"][0], intended_type)
        assert isinstance(new_array[2][0]["y"][1], intended_type)
        assert isinstance(new_array[2][0]["y"][3], intended_type)
        assert isinstance(new_array[2][1], intended_type)
        assert isinstance(new_array[2][2], intended_type)
        assert isinstance(new_array[2][3], intended_type)
        assert isinstance(new_array[2][4]["x"], intended_type)
        assert isinstance(new_array[2][4]["y"][0], intended_type)
        assert isinstance(new_array[2][4]["y"][1], intended_type)
        assert isinstance(new_array[2][4]["y"][3], intended_type)
        assert isinstance(new_array[2][4]["y"][4], intended_type)

    int_type32 = ak.operations.full_like(array, 12, dtype="int32")
    int_type64 = ak.operations.full_like(array, 12, dtype="int64")
    float_type = ak.operations.full_like(array, 12, dtype=float)

    assert int_type64.to_list() == float_type.to_list()
    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)

    bool_type = ak.operations.full_like(array, 12, dtype=bool)
    assert_array_type(bool_type, np.bool_)

    int_type32 = ak.operations.full_like(array, -1.2, dtype="int32")
    int_type64 = ak.operations.full_like(array, -1.2, dtype="int64")
    float_type = ak.operations.full_like(array, -1.2, dtype=float)
    bool_type = ak.operations.full_like(array, -1.2, dtype=bool)

    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)
    assert_array_type(bool_type, np.bool_)

    int_type32 = ak.operations.zeros_like(array, dtype="int32")
    int_type64 = ak.operations.zeros_like(array, dtype="int64")
    float_type = ak.operations.zeros_like(array, dtype=float)
    bool_type = ak.operations.zeros_like(array, dtype=bool)

    assert int_type64.to_list() == float_type.to_list()
    assert int_type64.to_list() == bool_type.to_list()
    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)
    assert_array_type(bool_type, np.bool_)

    int_type32 = ak.operations.ones_like(array, dtype="int32")
    int_type64 = ak.operations.ones_like(array, dtype="int64")
    float_type = ak.operations.ones_like(array, dtype=float)
    bool_type = ak.operations.ones_like(array, dtype=bool)

    assert int_type64.to_list() == float_type.to_list()
    assert int_type64.to_list() == bool_type.to_list()
    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)
    assert_array_type(bool_type, np.bool_)

    array = ak.Array([["one", "two", "three"], [], ["four", "five"]])

    def assert_array_type(new_array, intended_type):
        """helper function to check each part of the array took the intended type"""
        assert isinstance(new_array[0][0], intended_type)
        assert isinstance(new_array[0][1], intended_type)
        assert isinstance(new_array[0][2], intended_type)
        assert isinstance(new_array[2][0], intended_type)
        assert isinstance(new_array[2][1], intended_type)

    int_type32 = ak.operations.full_like(array, 12, dtype="int32")
    int_type64 = ak.operations.full_like(array, 12, dtype="int64")
    float_type = ak.operations.full_like(array, 12, dtype=float)

    assert int_type64.to_list() == float_type.to_list()
    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)

    bool_type = ak.operations.full_like(array, 12, dtype=bool)
    assert_array_type(bool_type, np.bool_)

    int_type32 = ak.operations.full_like(array, -1.2, dtype="int32")
    int_type64 = ak.operations.full_like(array, -1.2, dtype="int64")
    float_type = ak.operations.full_like(array, -1.2, dtype=float)
    bool_type = ak.operations.full_like(array, -1.2, dtype=bool)

    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)
    assert_array_type(bool_type, np.bool_)

    int_type32 = ak.operations.zeros_like(array, dtype="int32")
    int_type64 = ak.operations.zeros_like(array, dtype="int64")
    float_type = ak.operations.zeros_like(array, dtype=float)
    bool_type = ak.operations.zeros_like(array, dtype=bool)

    assert int_type64.to_list() == float_type.to_list()
    assert int_type64.to_list() == bool_type.to_list()
    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)
    assert_array_type(bool_type, np.bool_)

    int_type32 = ak.operations.ones_like(array, dtype="int32")
    int_type64 = ak.operations.ones_like(array, dtype="int64")
    float_type = ak.operations.ones_like(array, dtype=float)
    bool_type = ak.operations.ones_like(array, dtype=bool)
    assert int_type64.to_list() == float_type.to_list()
    assert int_type64.to_list() == bool_type.to_list()
    assert_array_type(int_type32, np.int32)
    assert_array_type(int_type64, np.int64)
    assert_array_type(float_type, float)
    assert_array_type(bool_type, np.bool_)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak

ROOT = pytest.importorskip("ROOT")

ROOT.ROOT.EnableImplicitMT(1)

compiler = ROOT.gInterpreter.Declare


def test_data_frame_integers():
    ak_array_x = ak.Array([1, 2, 3, 4, 5])
    ak_array_y = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])

    data_frame = ak.to_rdataframe({"x": ak_array_x, "y": ak_array_y})

    assert data_frame.GetColumnType("x") == "int64_t"
    assert data_frame.GetColumnType("y") == "double"

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x", "y"),
    )
    assert ak_array_x.to_list() == ak_array_out["x"].to_list()
    assert ak_array_y.to_list() == ak_array_out["y"].to_list()


def test_data_frame_vec_of_vec_of_real():
    ak_array_in = ak.Array([[[1.1], [2.2]], [[3.3], [4.4, 5.5]]])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x").startswith("awkward::ListArray_")

    ak_array_out = ak.from_rdataframe(
        data_frame,
        columns=("x",),
    )
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_filter():
    ak_array_x = ak.Array([1, 2, 3, 4, 5])
    ak_array_y = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])

    data_frame = ak.to_rdataframe({"x": ak_array_x, "y": ak_array_y})
    rdf3 = data_frame.Filter("x > 3")

    assert data_frame.GetColumnType("x") == "int64_t"
    assert data_frame.GetColumnType("y") == "double"

    ak_array_out = ak.from_rdataframe(
        rdf3,
        columns=(
            "x",
            "y",
        ),
    )
    assert ak_array_x[3:].to_list() == ak_array_out["x"].to_list()
    assert ak_array_y[3:].to_list() == ak_array_out["y"].to_list()


def test_data_frame_rvec_filter():
    ak_array_x = ak.Array([[1, 2], [3], [4, 5]])
    ak_array_y = ak.Array([[1.0, 1.1], [2.2, 3.3, 4.4], [5.5]])

    data_frame = ak.to_rdataframe({"x": ak_array_x, "y": ak_array_y})
    rdf3 = data_frame.Filter("x.size() >= 2")

    assert data_frame.GetColumnType("x") == "ROOT::VecOps::RVec<int64_t>"
    assert data_frame.GetColumnType("y") == "ROOT::VecOps::RVec<double>"

    ak_array_out = ak.from_rdataframe(
        rdf3,
        columns=(
            "x",
            "y",
        ),
    )
    assert ak_array_out["x"].to_list() == [[1, 2], [4, 5]]
    assert ak_array_out["y"].to_list() == [[1.0, 1.1], [5.5]]

    rdf4 = data_frame.Filter("y.size() == 2")
    ak_array_out = ak.from_rdataframe(
        rdf4,
        columns=(
            "x",
            "y",
        ),
    )
    assert ak_array_out["x"].to_list() == [[1, 2]]
    assert ak_array_out["y"].to_list() == [[1.0, 1.1]]


def test_data_frame_double():
    ak_array_in = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])

    data_frame = ak.to_rdataframe({"x": ak_array_in})

    assert data_frame.GetColumnType("x") == "double"

    ak_array_out = ak.from_rdataframe(data_frame, columns=("x",))
    assert ak_array_in.to_list() == ak_array_out["x"].to_list()


def test_data_frame_vec_of_vec():
    array = ak.Array(
        [
            [
                {"x": 1.1, "y": [1]},
                {"x": None, "y": [1, 2]},
                {"x": 3.3, "y": [1, 2, 3]},
            ],
            [],
            [{"x": None, "y": [1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}],
        ]
    )
    # ] * 10000)

    rdf2 = ak.to_rdataframe({"array": array})
    # We create a matrix RxC here
    # Note when dimensions R and C are large, the following code suffers
    # from potential performance penalties caused by frequent reallocation
    # of memory by the push_back() function. This should be used only when
    # vector dimensions are not known in advance.
    rdf3 = rdf2.Define(
        "output",
        """
    std::vector<std::vector<double>> tmp1;

    for (auto record : array) {
        std::vector<double> tmp2;
        for (auto number : record.y()) {
            tmp2.push_back(number * number);
        }
        tmp1.push_back(tmp2);
    }
    return tmp1;
    """,
    )

    assert rdf3.GetColumnType("output") == "vector<vector<double> >"

    rdf4 = rdf3.Define(
        "output2",
        """
    std::vector<std::vector<std::vector<double>>> tmp1;

    for (auto record : array) {
        std::vector<std::vector<double>> tmp2;
        // we can check if it's None:
        // if (record.x().has_value())
        // or set it to 1 so that we do not scale:
        double x_number = record.x().value_or(1);
        for (auto number : record.y()) {
            std::vector<double> tmp3;
            for (int64_t i = 0; i < std::rint(x_number); i++) {
                double value = x_number * number;
                tmp3.push_back(value);
            }
            tmp2.push_back(tmp3);
        }
        tmp1.push_back(tmp2);
    }
    return tmp1;
    """,
    )
    assert rdf4.GetColumnType("output2") == "vector<vector<vector<double> > >"

    out = ak.from_rdataframe(
        rdf4,
        columns=(
            "output",
            "output2",
        ),
    )

    result = ak.Array(
        [
            [
                [[1.1]],  # "x" is 1 - "y" values are unchanged, and each is nesed
                [
                    [1.0],
                    [2.0],
                ],  # "x" is None - "y" values are unchanged, and each is nesed
                [
                    [3.3, 3.3, 3.3],
                    [6.6, 6.6, 6.6],
                    [9.899999999999999, 9.899999999999999, 9.899999999999999],
                ],  # "x" is 3.3 - "y" values are scaled by 3.3 and each is nesed 3 times
            ],
            [],
            [
                [
                    [1.0],
                    [2.0],
                    [3.0],
                    [4.0],
                ],  # "x" is None - "y" values are unchanged, and each is nesed
                [
                    [5.5, 5.5, 5.5, 5.5, 5.5, 5.5],
                    [11.0, 11.0, 11.0, 11.0, 11.0, 11.0],
                    [16.5, 16.5, 16.5, 16.5, 16.5, 16.5],
                    [22.0, 22.0, 22.0, 22.0, 22.0, 22.0],
                    [27.5, 27.5, 27.5, 27.5, 27.5, 27.5],
                ],  # "x" is 5.5 - "y" values are scaled by 5.5 and each is nesed 5 times
            ],
        ]
    )
    assert out["output2"].to_list() == result.to_list()


def test_rdata_frame_rvecs_as_records():
    data_frame = ROOT.RDataFrame(1024)
    coordDefineCode = """ROOT::VecOps::RVec<double> {0}(len);
                     std::transform({0}.begin(), {0}.end(), {0}.begin(), [](double){{return gRandom->Uniform(-1.0, 1.0);}});
                     return {0};"""

    data_frame_x_y = (
        data_frame.Define("len", "gRandom->Uniform(0, 16)")
        .Define("x", coordDefineCode.format("x"))
        .Define("y", coordDefineCode.format("y"))
    )

    # Now we have in hands d, a RDataFrame with two columns, x and y, which
    # hold collections of coordinates. The size of these collections vary.
    # Let's now define radii out of x and y. We'll do it treating the collections
    # stored in the columns without looping on the individual elements.
    data_frame_x_y_r = data_frame_x_y.Define("r", "sqrt(x*x + y*y)")
    assert data_frame_x_y_r.GetColumnType("r") == "ROOT::VecOps::RVec<double>"

    array = ak.from_rdataframe(
        data_frame_x_y_r,
        columns=(
            "x",
            "y",
            "r",
        ),
    )

    assert array["x"].layout.form == ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64")
    )
    assert array["y"].layout.form == ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64")
    )
    assert array["r"].layout.form == ak.forms.ListOffsetForm(
        "i64", ak.forms.NumpyForm("float64")
    )

    assert array.layout.form == ak.forms.RecordForm(
        [
            ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64")),
            ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64")),
            ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64")),
        ],
        ["x", "y", "r"],
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import sys

import awkward as ak
from awkward._nplikes.shape import unknown_length
from awkward._regularize import is_integer


class ArrayType:
    def __init__(self, content, length, behavior=None):
        if not isinstance(content, ak.types.Type):
            raise TypeError(
                "{} all 'contents' must be Type subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not ((is_integer(length) and length >= 0) or length is unknown_length):
            raise ValueError(
                "{} 'length' must be a non-negative integer or unknown length, not {}".format(
                    type(self).__name__, repr(length)
                )
            )
        self._content = content
        self._length = length
        self._behavior = behavior

    @property
    def content(self):
        return self._content

    @property
    def length(self):
        return self._length

    @property
    def behavior(self):
        return self._behavior

    def __str__(self):
        return "".join(self._str("", True))

    def show(self, stream=sys.stdout):
        stream.write("".join([*self._str("", False), "\n"]))

    def _str(self, indent, compact):
        return [
            f"{self._length} * ",
            *self._content._str(
                indent,
                compact,
                self._behavior,
            ),
        ]

    def __repr__(self):
        args = [repr(self._content), repr(self._length), repr(self._behavior)]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def is_equal_to(self, other, *, all_parameters: bool = False) -> bool:
        return (
            isinstance(other, type(self))
            and (
                other._length is unknown_length
                or self._length is unknown_length
                or self._length == other._length
            )
            and self._content.is_equal_to(other._content, all_parameters=all_parameters)
        )

    __eq__ = is_equal_to



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("softmax",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._nplikes import ufuncs
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def softmax(x, axis=None, *, keepdims=False, mask_identity=False):
    """
    Args:
        x: The data on which to compute the softmax (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this function decreases the number of
            dimensions by 1; if True, the output values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, the application of this function on
            empty lists results in None (an option type); otherwise, the
            calculation is followed through with the reducers' identities,
            usually resulting in floating-point `nan`.

    Computes the softmax in each group of elements from `x` (many
    types supported, including all Awkward Arrays and Records). The grouping
    is performed the same way as for reducers, though this operation is not a
    reducer and has no identity.

    This function has no NumPy equivalent.

    Passing all arguments to the reducers, the softmax is calculated as

        np.exp(x) / ak.sum(np.exp(x))

    See #ak.sum for a complete description of handling nested lists and
    missing values (None) in reducers, and #ak.mean for an example with another
    non-reducer.
    """
    # Dispatch
    yield (x,)

    # Implementation
    return _impl(x, axis, keepdims, mask_identity)


def _impl(x, axis, keepdims, mask_identity):
    axis = regularize_axis(axis)
    behavior = behavior_of(x)
    x = ak.highlevel.Array(
        ak.operations.to_layout(x, allow_record=False, allow_other=False),
        behavior=behavior,
    )

    with np.errstate(invalid="ignore", divide="ignore"):
        expx = ufuncs.exp(x)
        denom = ak.operations.ak_sum._impl(
            expx,
            axis,
            keepdims,
            mask_identity,
            highlevel=True,
            behavior=behavior,
        )
        return expx / denom



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    np_array = np.ma.MaskedArray(
        [[1, 2, 3], [4, 5, 6]], mask=[[False, True, False], [True, True, False]]
    )
    ak_array = ak.from_numpy(np_array)

    assert ak_array.to_list() == [[1, None, 3], [None, None, 6]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak


def test():
    behavior = {}

    @ak.behaviors.mixins.mixin_class(behavior)
    class Blah:
        @property
        def blah(self):
            return self["x"]

    a = ak.operations.zip({"x": [[1, 2], [3]], "y": [[4, 5], [6]]})
    b = ak.operations.zip({"x": [[-1, -2, -3], [-4]], "z": [[-4, -5, -6], [-7]]})

    a2 = a[a.x % 2 == 0]
    b2 = b[b.x % 2 == 0]
    c2 = ak.operations.with_name(
        ak.operations.concatenate([a2, b2], axis=1), "Blah", behavior=behavior
    )

    assert c2.blah.to_list() == [[2, -2], [-4]]



from collections.abc import Callable, Collection, Generator
from functools import wraps
from inspect import isgenerator

from awkward._errors import OperationErrorContext
from awkward._typing import Any, TypeAlias, TypeVar

T = TypeVar("T")
DispatcherType: TypeAlias = "Callable[..., Generator[Collection[Any], None, T]]"
HighLevelType: TypeAlias = "Callable[..., T]"


def high_level_function(func: DispatcherType) -> HighLevelType:
    """Decorate a high-level function such that it may be overloaded by third-party array objects"""

    @wraps(func)
    def dispatch(*args, **kwargs):
        # NOTE: this decorator assumes that the operation is exposed under `ak.`
        with OperationErrorContext(f"ak.{func.__qualname__}", args, kwargs):
            gen_or_result = func(*args, **kwargs)
            if isgenerator(gen_or_result):
                array_likes = next(gen_or_result)
                assert isinstance(array_likes, Collection)

                # Permit a third-party array object to intercept the invocation
                for array_like in array_likes:
                    try:
                        custom_impl = array_like.__awkward_function__
                    except AttributeError:
                        continue
                    else:
                        result = custom_impl(dispatch, array_likes, args, kwargs)

                        # Future proof the implementation by permitting the `__awkward_function__` to return `NotImplemented`
                        # This may later be used to signal that another overload should be used.
                        if result is NotImplemented:
                            raise NotImplementedError
                        else:
                            return result

                # Failed to find a custom overload, so resume the original function
                try:
                    next(gen_or_result)
                except StopIteration as err:
                    return err.value
                else:
                    raise AssertionError(
                        "high-level functions should only implement a single yield statement"
                    )

            return gen_or_result

    return dispatch



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import collections
import functools
import inspect
from collections.abc import Iterable
from itertools import chain

import numpy

import awkward as ak
from awkward._backends.backend import Backend
from awkward._backends.dispatch import backend_of, common_backend
from awkward._behavior import (
    behavior_of,
    find_custom_cast,
    find_ufunc,
    find_ufunc_generic,
)
from awkward._categorical import as_hashable
from awkward._layout import wrap_layout
from awkward._nplikes import to_nplike
from awkward._parameters import parameters_intersect
from awkward._regularize import is_non_string_like_iterable
from awkward._typing import Any, Iterator, Mapping
from awkward._util import Sentinel
from awkward.contents.numpyarray import NumpyArray

UNSUPPORTED = Sentinel("UNSUPPORTED", __name__)


def convert_to_array(layout, args, kwargs):
    out = ak.operations.to_numpy(layout, allow_missing=False)
    if args == () and kwargs == {}:
        return out
    else:
        return numpy.array(out, *args, **kwargs)


implemented = {}


def _find_backends(args: Iterable) -> Iterator[Backend]:
    """
    Args:
        args: iterable of objects to visit

    Yields the encountered backends of layout / array-like arguments encountered
    in the argument list.
    """
    stack = collections.deque(args)
    while stack:
        arg = stack.popleft()
        # If the argument declares a backend, easy!
        backend = backend_of(arg, default=None)
        if backend is not None:
            yield backend
        # Otherwise, traverse into supported sequence types
        elif isinstance(arg, (tuple, list)):
            stack.extend(arg)


def _to_rectilinear(arg, backend: Backend):
    # Is this object something we already associate with a backend?
    arg_backend = backend_of(arg, default=None)
    if arg_backend is not None:
        arg_nplike = arg_backend.nplike
        # Is this argument already in a backend-supported form?
        if arg_nplike.is_own_array(arg):
            # Convert to the appropriate nplike
            return to_nplike(
                arg_nplike.asarray(arg), backend.nplike, from_nplike=arg_nplike
            )
        # Otherwise, cast to layout and convert
        else:
            layout = ak.to_layout(arg, allow_record=False, allow_other=False)
            return layout.to_backend(backend).to_backend_array(allow_missing=True)
    elif isinstance(arg, tuple):
        return tuple(_to_rectilinear(x, backend) for x in arg)
    elif isinstance(arg, list):
        return [_to_rectilinear(x, backend) for x in arg]
    elif is_non_string_like_iterable(arg):
        raise TypeError(
            f"encountered an unsupported iterable value {arg!r} whilst converting arguments to NumPy-friendly "
            f"types. If this argument should be supported, please file a bug report."
        )
    else:
        return arg


def array_function(func, types, args, kwargs: dict[str, Any], behavior: Mapping | None):
    function = implemented.get(func)
    if function is not None:
        return function(*args, **kwargs)
    # Use NumPy's implementation
    else:
        backend = common_backend(_find_backends(chain(args, kwargs.values())))

        rectilinear_args = tuple(_to_rectilinear(x, backend) for x in args)
        rectilinear_kwargs = {k: _to_rectilinear(v, backend) for k, v in kwargs.items()}
        result = func(*rectilinear_args, **rectilinear_kwargs)
        # We want the result to be a layout (this will fail for functions returning non-array convertibles)
        out = ak.operations.ak_to_layout._impl(
            result, allow_record=True, allow_other=True, regulararray=True
        )
        return wrap_layout(out, behavior=behavior, allow_other=True)


def implements(numpy_function):
    def decorator(function):
        signature = inspect.signature(function)
        unsupported_names = {
            p.name for p in signature.parameters.values() if p.default is UNSUPPORTED
        }

        @functools.wraps(function)
        def ensure_valid_args(*args, **kwargs):
            parameters = signature.bind(*args, **kwargs)
            provided_invalid_names = parameters.arguments.keys() & unsupported_names
            if provided_invalid_names:
                names = ", ".join(provided_invalid_names)
                raise TypeError(
                    f"Awkward NEP-18 overload was provided with unsupported argument(s): {names}"
                )
            return function(*args, **kwargs)

        implemented[getattr(numpy, numpy_function)] = ensure_valid_args
        return function

    return decorator


def _array_ufunc_custom_cast(inputs, behavior: Mapping | None, backend):
    args = [
        wrap_layout(x, behavior)
        if isinstance(x, (ak.contents.Content, ak.record.Record))
        else x
        for x in inputs
    ]

    nextinputs = []
    for x in args:
        cast_fcn = find_custom_cast(x, behavior)
        if cast_fcn is not None:
            x = cast_fcn(x)
        # String conversion
        elif isinstance(x, (str, bytes)):
            x = ak.to_layout([x])
        maybe_layout = ak.operations.to_layout(x, allow_record=True, allow_other=True)
        if isinstance(maybe_layout, (ak.contents.Content, ak.record.Record)):
            maybe_layout = maybe_layout.to_backend(backend)

        nextinputs.append(maybe_layout)
    return nextinputs


def _array_ufunc_adjust(
    custom, inputs, kwargs: dict[str, Any], behavior: Mapping | None
):
    args = [
        wrap_layout(x, behavior)
        if isinstance(x, (ak.contents.Content, ak.record.Record))
        else x
        for x in inputs
    ]
    out = custom(*args, **kwargs)
    if not isinstance(out, tuple):
        out = (out,)

    return tuple(
        x.layout if isinstance(x, (ak.highlevel.Array, ak.highlevel.Record)) else x
        for x in out
    )


def _array_ufunc_adjust_apply(
    apply_ufunc, ufunc, method, inputs, kwargs: dict[str, Any], behavior: Mapping | None
):
    nextinputs = [wrap_layout(x, behavior, allow_other=True) for x in inputs]
    out = apply_ufunc(ufunc, method, nextinputs, kwargs)

    if out is NotImplemented:
        return None
    else:
        if not isinstance(out, tuple):
            out = (out,)
        return tuple(
            x.layout if isinstance(x, (ak.highlevel.Array, ak.highlevel.Record)) else x
            for x in out
        )


def _array_ufunc_signature(ufunc, inputs):
    signature = [ufunc]
    for x in inputs:
        if isinstance(x, ak.contents.Content):
            record_name, list_name = x.parameter("__record__"), x.parameter("__list__")
            if record_name is not None:
                signature.append(record_name)
            elif list_name is not None:
                signature.append(list_name)
            elif isinstance(x, NumpyArray):
                signature.append(x.dtype.type)
            else:
                signature.append(None)
        else:
            signature.append(type(x))

    return tuple(signature)


def _array_ufunc_categorical(
    ufunc, method: str, inputs, kwargs: dict[str, Any], behavior: Mapping | None
):
    if (
        ufunc is numpy.equal
        and len(inputs) == 2
        and isinstance(inputs[0], ak.contents.Content)
        and inputs[0].is_indexed
        and inputs[0].parameter("__array__") == "categorical"
        and isinstance(inputs[1], ak.contents.Content)
        and inputs[1].is_indexed
        and inputs[1].parameter("__array__") == "categorical"
    ):
        assert method == "__call__"
        one, two = inputs

        one_index = numpy.asarray(one.index)
        two_index = numpy.asarray(two.index)
        one_content = wrap_layout(one.content, behavior)
        two_content = wrap_layout(two.content, behavior)

        if len(one_content) == len(two_content) and ak.operations.all(
            one_content == two_content, axis=None
        ):
            one_mapped = one_index

        else:
            one_list = ak.operations.to_list(one_content)
            two_list = ak.operations.to_list(two_content)
            one_hashable = [as_hashable(x) for x in one_list]
            two_hashable = [as_hashable(x) for x in two_list]
            two_lookup = {x: i for i, x in enumerate(two_hashable)}

            one_to_two = numpy.empty(len(one_hashable) + 1, dtype=numpy.int64)
            for i, x in enumerate(one_hashable):
                one_to_two[i] = two_lookup.get(x, len(two_hashable))
            one_to_two[-1] = -1

            one_mapped = one_to_two[one_index]

        out = one_mapped == two_index
        return (ak.contents.NumpyArray(out),)
    else:
        nextinputs = []
        for x in inputs:
            if isinstance(x, ak.contents.Content) and x.is_indexed:
                nextinputs.append(wrap_layout(x.project(), behavior=behavior))
            else:
                nextinputs.append(wrap_layout(x, behavior=behavior, allow_other=True))

        out = getattr(ufunc, method)(*nextinputs, **kwargs)
        if not isinstance(out, tuple):
            out = (out,)
        return tuple(ak.to_layout(x, allow_other=True) for x in out)


def _array_ufunc_string_likes(
    ufunc, method: str, inputs, kwargs: dict[str, Any], behavior: Mapping | None
):
    assert method == "__call__"

    if (
        ufunc in (numpy.equal, numpy.not_equal)
        and len(inputs) == 2
        and isinstance(inputs[0], ak.contents.Content)
        and isinstance(inputs[1], ak.contents.Content)
        and inputs[0].parameter("__array__") in ("string", "bytestring")
        and inputs[1].parameter("__array__") == inputs[0].parameter("__array__")
    ):
        left, right = inputs
        nplike = left.backend.nplike

        left = ak.without_parameters(left, highlevel=False)
        right = ak.without_parameters(right, highlevel=False)

        # first condition: string lengths must be the same
        counts1 = nplike.asarray(
            ak._do.reduce(left, ak._reducers.Count(), axis=-1, mask=False)
        )
        counts2 = nplike.asarray(
            ak._do.reduce(right, ak._reducers.Count(), axis=-1, mask=False)
        )

        out = counts1 == counts2

        # only compare characters in strings that are possibly equal (same length)
        possible = nplike.logical_and(out, counts1)
        possible_counts = counts1[possible]

        if len(possible_counts) > 0:
            onepossible = left[possible]
            twopossible = right[possible]
            reduced = ak.operations.all(
                wrap_layout(onepossible) == wrap_layout(twopossible),
                axis=-1,
                highlevel=False,
            )
            # update same-length strings with a verdict about their characters
            out[possible] = reduced.data

        if ufunc is numpy.not_equal:
            out = nplike.logical_not(out)

        return (ak.contents.NumpyArray(out),)


def array_ufunc(ufunc, method: str, inputs, kwargs: dict[str, Any]):
    if method != "__call__" or len(inputs) == 0 or "out" in kwargs:
        return NotImplemented

    behavior = behavior_of(*inputs)
    backend = backend_of(*inputs)

    inputs = _array_ufunc_custom_cast(inputs, behavior, backend)

    def action(inputs, **ignore):
        contents = [x for x in inputs if isinstance(x, ak.contents.Content)]

        signature = _array_ufunc_signature(ufunc, inputs)
        # Do we have a custom ufunc (an override of the given ufunc)?
        custom = find_ufunc(behavior, signature)
        if custom is not None:
            return _array_ufunc_adjust(custom, inputs, kwargs, behavior)

        # Do we have any categoricals?
        if any(
            x.is_indexed and x.parameter("__array__") == "categorical" for x in contents
        ):
            out = _array_ufunc_categorical(ufunc, method, inputs, kwargs, behavior)
            if out is not None:
                return out

        # Do we have any strings?
        if any(
            x.is_list and x.parameter("__array__") in ("string", "bytestring")
            for x in contents
        ):
            out = _array_ufunc_string_likes(ufunc, method, inputs, kwargs, behavior)
            if out is not None:
                return out

            # Do we have all-strings? If so, we can't proceed
            if all(
                x.is_list and x.parameter("__array__") in ("string", "bytestring")
                for x in contents
            ):
                raise TypeError(
                    "{}.{} is not implemented for string types. "
                    "To register an implementation, add a name to these string(s) and register a behavior overload".format(
                        type(ufunc).__module__, ufunc.__name__
                    )
                )

        if ufunc is numpy.matmul:
            raise NotImplementedError(
                "matrix multiplication (`@` or `np.matmul`) is not yet implemented for Awkward Arrays"
            )

        # Do we have a custom generic ufunc override (a function that accepts _all_ ufuncs)?
        for x in contents:
            apply_ufunc = find_ufunc_generic(ufunc, x, behavior)
            if apply_ufunc is not None:
                out = _array_ufunc_adjust_apply(
                    apply_ufunc, ufunc, method, inputs, kwargs, behavior
                )
                if out is not None:
                    return out

        if all(
            isinstance(x, NumpyArray) or not isinstance(x, ak.contents.Content)
            for x in inputs
        ):
            nplike = backend.nplike

            # Broadcast parameters against one another
            parameters = functools.reduce(
                parameters_intersect, [x._parameters for x in contents]
            )

            args = []
            for x in inputs:
                if isinstance(x, NumpyArray):
                    args.append(x._raw(nplike))
                else:
                    args.append(x)

            # Give backend a chance to change the ufunc implementation
            impl = backend.prepare_ufunc(ufunc)

            # Invoke ufunc
            result = impl(*args, **kwargs)

            return (NumpyArray(result, backend=backend, parameters=parameters),)

        # Do we have exclusively nominal types without custom overloads?
        if all(
            x.parameter("__list__") is not None or x.parameter("__record__") is not None
            for x in contents
        ):
            error_message = []
            for x in inputs:
                if isinstance(x, ak.contents.Content):
                    if x.parameter("__list__") is not None:
                        error_message.append(x.parameter("__list__"))
                    elif x.parameter("__record__") is not None:
                        error_message.append(x.parameter("__record__"))
                    else:
                        error_message.append(type(x).__name__)
                else:
                    error_message.append(type(x).__name__)
            raise TypeError(
                "no {}.{} overloads for custom types: {}".format(
                    type(ufunc).__module__, ufunc.__name__, ", ".join(error_message)
                )
            )

        return None

    if sum(int(isinstance(x, ak.contents.Content)) for x in inputs) == 1:
        where = None
        for i, x in enumerate(inputs):
            if isinstance(x, ak.contents.Content):
                where = i
                break
        assert where is not None

        nextinputs = list(inputs)

        def unary_action(layout, **ignore):
            nextinputs[where] = layout
            result = action(tuple(nextinputs), **ignore)
            if result is None:
                return None
            else:
                assert isinstance(result, tuple) and len(result) == 1
                return result[0]

        out = ak._do.recursively_apply(
            inputs[where],
            unary_action,
            behavior,
            function_name=ufunc.__name__,
            allow_records=False,
        )

    else:
        out = ak._broadcasting.broadcast_and_apply(
            inputs, action, behavior, allow_records=False, function_name=ufunc.__name__
        )
        assert isinstance(out, tuple) and len(out) == 1
        out = out[0]

    return wrap_layout(out, behavior)


def action_for_matmul(inputs):
    raise NotImplementedError
